from abc import ABC
from collections.abc import Callable, Iterable
from dataclasses import dataclass
from math import prod
from typing import ClassVar, TypeVar, cast

from xdsl.context import Context
from xdsl.dialects import arith, builtin, func, memref, mpi, printf, scf, stencil
from xdsl.dialects.builtin import ContainerType
from xdsl.dialects.experimental import dmp
from xdsl.ir import Attribute, Block, Operation, OpResult, Region, SSAValue
from xdsl.passes import ModulePass
from xdsl.pattern_rewriter import (
    GreedyRewritePatternApplier,
    PatternRewriter,
    PatternRewriteWalker,
    RewritePattern,
    op_type_rewrite_pattern,
)
from xdsl.rewriter import InsertPoint, Rewriter
from xdsl.transforms.experimental.convert_stencil_to_ll_mlir import StencilToMemRefType

_T = TypeVar("_T", bound=Attribute)

_rank_dtype = builtin.i32


@dataclass
class ChangeStoreOpSizes(RewritePattern):
    strategy: dmp.DomainDecompositionStrategy

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: stencil.StoreOp, rewriter: PatternRewriter, /):
        assert all(
            integer_attr.data == 0 for integer_attr in op.bounds.lb.array.data
        ), "lb must be 0"
        shape: tuple[int, ...] = tuple(
            integer_attr.data for integer_attr in op.bounds.ub.array.data
        )
        new_shape = self.strategy.calc_resize(shape)
        op.bounds = stencil.StencilBoundsAttr.new(
            [
                stencil.IndexAttr.get(*(len(new_shape) * [0])),
                stencil.IndexAttr.get(*new_shape),
            ]
        )


@dataclass
class AddHaloExchangeOps(RewritePattern):
    """
    This rewrite adds a `stencil.halo_exchange` after each `stencil.load` op
    """

    strategy: dmp.DomainDecompositionStrategy

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: stencil.LoadOp, rewriter: PatternRewriter, /):
        swap_op = dmp.SwapOp.get(op.res, self.strategy)
        assert swap_op.swapped_values
        rewriter.insert_op_after_matched_op(swap_op)
        for use in op.res.uses.copy():
            if use.operation is swap_op:
                continue
            use.operation.operands[use.index] = swap_op.swapped_values
            rewriter.handle_operation_modification(use.operation)


@dataclass
class LowerHaloExchangeToMpi(RewritePattern):
    init: bool
    debug_prints: bool = False

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: dmp.SwapOp, rewriter: PatternRewriter, /):
        exchanges = list(op.swaps)

        input_type = cast(ContainerType[Attribute], op.input_stencil.type)

        rewriter.replace_matched_op(
            list(
                generate_mpi_calls_for(
                    op.input_stencil,
                    exchanges,
                    input_type.get_element_type(),
                    op.strategy.comm_layout(),
                    emit_init=self.init,
                    emit_debug=self.debug_prints,
                )
            ),
            [],
        )


def _generate_single_axis_calc_and_check(
    pos_in_axis: SSAValue,
    offset_in_axis: int,
    axis_size: int,
) -> tuple[list[Operation], SSAValue, SSAValue]:
    """
    Given a position (in SSA), aand compile time known offset/size, generate the
    following operations:

    dest = pos_in_axis + offset_in_axis
    is_valid_axsis = (0 <= dest && dest < axis_size)

    Since we know the offset at compile time, we can skip one of the calculations.

    Returns the ops, and the ssa values that contain dest and is_valid_axis
    Return is ([ops], dest, is_valid)
    """
    # if the offset is zero, we can skip the comparison and just return true
    # because my_pos + 0 is always inbounds!
    if offset_in_axis == 0:
        return (
            [true := arith.ConstantOp.from_int_and_width(1, 1)],
            pos_in_axis,
            true.result,
        )

    # check if we are decrementing or increment the position here
    is_decrement = offset_in_axis < 0
    # very important that we use signed arithmetic here!
    # find the correct comparison for
    # is_valid_axsis = (0 <= dest && dest < axis_size)
    # since we know if we will be incrementing or decrementing, we can skip one of the
    # checks
    comparison = "sge" if is_decrement else "slt"

    return (
        [
            offset_v := arith.ConstantOp.from_int_and_width(
                offset_in_axis, _rank_dtype
            ),
            dest := arith.AddiOp(pos_in_axis, offset_v),
            # get the bound we need to check:
            bound := arith.ConstantOp.from_int_and_width(
                0 if is_decrement else axis_size, _rank_dtype
            ),
            # comparison == true <=> we have a valid dest positon
            cond_val := arith.CmpiOp(dest, bound, comparison),
        ],
        dest.result,
        cond_val.result,
    )


def _grid_coords_from_rank(
    my_rank: SSAValue, grid: dmp.RankTopoAttr
) -> tuple[list[Operation], list[SSAValue]]:
    """
    Takes a rank and a dmp.grid, and returns operations to calculate
    the grid coordinates of the rank.
    """
    # a collection of all ops we want to return
    ret_ops: list[Operation] = []
    # the nodes coordinates in grid-space
    node_pos_nd: list[SSAValue] = []

    shape = grid.as_tuple()
    divisors = [prod(shape[i + 1 :]) for i in range(len(shape))]

    carry = my_rank
    for div in divisors:
        imm = arith.ConstantOp.from_int_and_width(div, builtin.i32)
        coord_i = arith.DivUIOp(carry, imm)
        carry = arith.RemUIOp(carry, imm)

        ret_ops.extend([imm, coord_i, carry])
        node_pos_nd.append(coord_i.result)

    return ret_ops, node_pos_nd


def _generate_dest_rank_computation(
    my_rank: SSAValue,
    offsets: tuple[int, ...],
    grid: dmp.RankTopoAttr,
) -> tuple[list[Operation], SSAValue, SSAValue]:
    """
    Takes the current rank, a tuple of offsets in grid coords, and a dmp.grid

    Calculates the dest rank, and comparisons if communication is in-bounds

    Returns a list of ops, the dest rank, and if it's in-bounds.

    Returns ([ops], dest_rank, is_in_bounds)
    """
    # calc grid coordinates
    ret_ops, node_pos_nd = _grid_coords_from_rank(my_rank, grid)

    # then we calculate the new coordinates:
    # save the condition vals somewhere
    condition_vals: list[SSAValue] = []
    # save the grid-coordinates of the destination rank
    dest_pos_nd: list[SSAValue] = []
    for pos_in_axis, offset_in_axis, axis_size in zip(
        node_pos_nd, offsets, grid.as_tuple()
    ):
        ops, dest, is_valid = _generate_single_axis_calc_and_check(
            pos_in_axis, offset_in_axis, axis_size
        )
        ret_ops.extend(ops)
        dest_pos_nd.append(dest)
        condition_vals.append(is_valid)

    # "and" all the condition vals
    accumulated_cond_val: SSAValue = condition_vals[0]
    for val in condition_vals[1:]:
        cmp = arith.AndIOp(accumulated_cond_val, val)
        ret_ops.append(cmp)
        accumulated_cond_val = cmp.result

    # calculate rank of destination node from grid coords

    carry: SSAValue = dest_pos_nd[-1]

    shape = grid.as_tuple()
    multiples = [prod(shape[i + 1 :]) for i in range(len(shape))]

    for pos, mul in zip(dest_pos_nd[:-1], multiples[:-1]):
        val = arith.ConstantOp.from_int_and_width(mul, builtin.i32)
        intermediate = arith.MuliOp(val, pos)
        carry_op = arith.AddiOp(carry, intermediate)
        carry = carry_op.result
        ret_ops.extend([val, intermediate, carry_op])

    return ret_ops, carry, accumulated_cond_val


def generate_mpi_calls_for(
    source: SSAValue,
    exchanges: list[dmp.ExchangeDeclarationAttr],
    dtype: Attribute,
    grid: dmp.RankTopoAttr,
    emit_init: bool = True,
    emit_debug: bool = False,
) -> Iterable[Operation]:
    # call mpi init (this will be hoisted to function level)
    if emit_init:
        yield mpi.InitOp()
    # allocate request array
    # we need two request objects per exchange
    # one for the send, one for the recv
    req_cnt = arith.ConstantOp.from_int_and_width(len(exchanges) * 2, builtin.i32)
    reqs = mpi.AllocateTypeOp(mpi.RequestType, req_cnt)
    # get comm rank
    rank = mpi.CommRankOp()
    # define static tag of 0
    tag = arith.ConstantOp.from_int_and_width(0, builtin.i32)

    yield from (req_cnt, reqs, rank, tag)

    recv_buffers: list[
        tuple[dmp.ExchangeDeclarationAttr, memref.AllocOp, SSAValue]
    ] = []

    for i, ex in enumerate(exchanges):
        # generate a temp buffer to store the data in
        reduced_size = [i for i in ex.size if i != 1]
        alloc_outbound = memref.AllocOp.get(dtype, 64, reduced_size)
        alloc_outbound.memref.name_hint = f"send_buff_ex{i}"
        alloc_inbound = memref.AllocOp.get(dtype, 64, reduced_size)
        alloc_inbound.memref.name_hint = f"recv_buff_ex{i}"
        yield from (alloc_outbound, alloc_inbound)

        # calc dest rank and check if it's in-bounds
        ops, dest_rank, is_in_bounds = _generate_dest_rank_computation(
            rank.rank, ex.neighbor, grid
        )
        yield from ops

        recv_buffers.append((ex, alloc_inbound, is_in_bounds))

        # get two unique indices
        cst_i = arith.ConstantOp.from_int_and_width(i, builtin.i32)
        cst_in = arith.ConstantOp.from_int_and_width(i + len(exchanges), builtin.i32)
        yield from (cst_i, cst_in)
        # from these indices, get request objects
        req_send = mpi.VectorGetOp(reqs, cst_i)
        req_recv = mpi.VectorGetOp(reqs, cst_in)
        yield from (req_send, req_recv)

        def then() -> Iterable[Operation]:
            # copy source area to outbound buffer
            yield from generate_memcpy(source, ex.source_area(), alloc_outbound.memref)
            # get ptr, count, dtype
            unwrap_out = mpi.UnwrapMemRefOp(alloc_outbound)
            unwrap_out.ptr.name_hint = f"send_buff_ex{i}_ptr"
            yield unwrap_out

            if emit_debug:
                yield printf.PrintFormatOp(
                    f"Rank {{}}: sending {ex.source_area()} -> {{}}", rank, dest_rank
                )

            # isend call
            yield mpi.IsendOp(
                unwrap_out.ptr,
                unwrap_out.len,
                unwrap_out.type,
                dest_rank,
                tag,
                req_send,
            )

            # get ptr for receive buffer
            unwrap_in = mpi.UnwrapMemRefOp(alloc_inbound)
            unwrap_in.ptr.name_hint = f"recv_buff_ex{i}_ptr"
            yield unwrap_in
            # Irecv call
            yield mpi.IrecvOp(
                unwrap_in.ptr,
                unwrap_in.len,
                unwrap_in.type,
                dest_rank,
                tag,
                req_recv,
            )
            yield scf.YieldOp()

        def else_() -> Iterable[Operation]:
            # set the request object to MPI_REQUEST_NULL s.t. they are ignored
            # in the waitall call
            yield mpi.NullRequestOp(req_send)
            yield mpi.NullRequestOp(req_recv)
            yield scf.YieldOp()

        yield scf.IfOp(
            is_in_bounds,
            [],
            Region([Block(then())]),
            Region([Block(else_())]),
        )

    # wait for all calls to complete
    yield mpi.WaitallOp(reqs.result, req_cnt.result)

    # start shuffling data into the main memref again
    for ex, buffer, cond_val in recv_buffers:
        yield scf.IfOp(
            cond_val,
            [],
            Region(
                [
                    Block(
                        list(
                            generate_memcpy(
                                source,
                                ex,
                                buffer.memref,
                                receive=True,
                            )
                        )
                        + [
                            printf.PrintFormatOp(
                                f"Rank {{}} receiving from {ex.neighbor}",
                                rank,
                            )
                        ]
                        * (1 if emit_debug else 0)
                        + [scf.YieldOp()]
                    )
                ]
            ),
            Region([Block([scf.YieldOp()])]),
        )


def generate_memcpy(
    field: SSAValue,
    ex: dmp.ExchangeDeclarationAttr,
    buffer: SSAValue,
    receive: bool = False,
) -> list[Operation]:
    """
    This function generates a memcpy routine to copy over the parts
    specified by the `field` from `field` into `buffer`.

    If receive=True, it instead copy from `buffer` into the parts of
    `field` as specified by `ex`

    """
    field_type = cast(stencil.FieldType[Attribute], field.type)
    assert isinstance(field_type.bounds, stencil.StencilBoundsAttr)
    memref_type = StencilToMemRefType(field_type)

    uc = builtin.UnrealizedConversionCastOp.get([field], result_type=[memref_type])

    memref_val = uc.results[0]

    offset = stencil.IndexAttr.get(*ex.offset) - field_type.bounds.lb

    subview = memref.SubviewOp.from_static_parameters(
        memref_val,
        memref_type,
        tuple(offset),
        ex.size,
        [1] * len(ex.offset),
        reduce_rank=True,
    )
    if receive:
        copy = memref.CopyOp(buffer, subview)
    else:
        copy = memref.CopyOp(subview, buffer)

    return [
        uc,
        subview,
        copy,
    ]


class MpiLoopInvariantCodeMotion:
    """
    THIS IS NOT A REWRITE PATTERN!

    This is a two-stage rewrite that modifies operations in a manner
    that is incompatible with the PatternRewriter!

    It implements a custom rewrite_module() method directly
    on the class.

    This rewrite moves all memref.allo, mpi.comm.rank, mpi.allocate
    and mpi.unwrap_memref ops and moves them "up" until it hits a
    func.func, and then places them *before* the op they appear in.
    """

    seen_ops: set[Operation]
    has_init: set[func.FuncOp]

    def __init__(self):
        self.seen_ops = set()
        self.has_init = set()

    def rewrite(
        self,
        op: (
            memref.AllocOp
            | mpi.CommRankOp
            | mpi.AllocateTypeOp
            | mpi.UnwrapMemRefOp
            | mpi.InitOp
        ),
        rewriter: Rewriter,
        /,
    ):
        if op in self.seen_ops:
            return
        self.seen_ops.add(op)

        # memref unwraps can always be moved to their allocation
        if isinstance(op, mpi.UnwrapMemRefOp) and isinstance(
            op.ref.owner, memref.AllocOp
        ):
            op.detach()
            rewriter.insert_op(op, InsertPoint.after(op.ref.owner))
            return

        base = op
        parent = op.parent_op()
        # walk upwards until we hit a function
        while parent is not None and not isinstance(parent, func.FuncOp):
            base = parent
            parent = base.parent_op()

        # check that we did not run into "nowhere"
        assert parent is not None, "Expected MPI to be inside a func.FuncOp!"
        assert isinstance(parent, func.FuncOp)  # this must be true now

        # check that we "ascended"
        if base == op:
            return

        if not can_loop_invariant_code_move(op):
            return

        # if we move an mpi.init, generate a finalize()!
        if isinstance(op, mpi.InitOp):
            # ignore multiple inits
            if parent in self.has_init:
                rewriter.erase_op(op)
                return
            self.has_init.add(parent)
            # add a finalize() call to the end of the function
            block = parent.regions[0].blocks[-1]
            return_op = block.last_op
            assert return_op is not None
            rewriter.insert_op(mpi.FinalizeOp(), InsertPoint.before(return_op))

        ops = list(collect_args_recursive(op))
        for found_op in ops:
            found_op.detach()
            rewriter.insert_op(found_op, InsertPoint.before(base))

    def get_matcher(
        self,
        worklist: list[
            memref.AllocOp
            | mpi.CommRankOp
            | mpi.AllocateTypeOp
            | mpi.UnwrapMemRefOp
            | mpi.InitOp
        ],
    ) -> Callable[[Operation], None]:
        """
        Returns a match() function that adds methods to a worklist
        if they satisfy some criteria.
        """

        def match(op: Operation):
            if isinstance(
                op,
                memref.AllocOp
                | mpi.CommRankOp
                | mpi.AllocateTypeOp
                | mpi.UnwrapMemRefOp
                | mpi.InitOp,
            ):
                worklist.append(op)

        return match

    def rewrite_module(self, op: builtin.ModuleOp):
        """
        Apply the rewrite to a module.

        We do a two-stage rewrite because we are modifying
        the operations we loop on them, which would throw of `op.walk`.
        """
        # collect all ops that should be rewritten
        worklist: list[
            memref.AllocOp
            | mpi.CommRankOp
            | mpi.AllocateTypeOp
            | mpi.UnwrapMemRefOp
            | mpi.InitOp
        ] = list()
        matcher = self.get_matcher(worklist)
        for o in op.walk():
            matcher(o)

        # rewrite ops
        rewriter = Rewriter()
        for matched_op in worklist:
            self.rewrite(matched_op, rewriter)


_LOOP_INVARIANT_OPS = (arith.ConstantOp, arith.AddiOp, arith.MuliOp)


def can_loop_invariant_code_move(op: Operation):
    """
    This function walks the def-use chain up to see if all the args are
    "constant enough" to move outside the loop.

    This check is very conservative, but that means it definitely works!
    """

    for arg in op.operands:
        if not isinstance(arg, OpResult):
            print(f"{arg} is not opresult")
            return False
        if not isinstance(arg.owner, _LOOP_INVARIANT_OPS):
            print(f"{arg} is not loop invariant")
            return False
        if not can_loop_invariant_code_move(arg.owner):
            return False
    return True


def collect_args_recursive(op: Operation) -> Iterable[Operation]:
    """
    Collect the def-use chain "upwards" of an operation.
    Check with can_loop_invariant_code_move prior to using this!
    """
    for arg in op.operands:
        assert isinstance(arg, OpResult)
        yield from collect_args_recursive(arg.owner)
    yield op


@dataclass(frozen=True)
class DmpDecompositionPass(ModulePass, ABC):
    """
    Represents a pass that takes a strategy as input
    """


@dataclass(frozen=True)
class DistributeStencilPass(DmpDecompositionPass):
    """
    Decompose a stencil to apply to a local domain.

    This pass applies stencil shape inference!
    """

    name = "distribute-stencil"

    STRATEGIES: ClassVar[dict[str, type[dmp.GridSlice2dAttr | dmp.GridSlice3dAttr]]] = {
        "2d-grid": dmp.GridSlice2dAttr,
        "3d-grid": dmp.GridSlice3dAttr,
    }

    slices: tuple[int, ...]
    """
    Number of slices to decompose the input into
    """

    strategy: str
    """
    Name of the decomposition strategy to use, see STRATEGIES property for options
    """

    restrict_domain: bool = True
    """
    Apply the domain restriction (i.e. change the stencil.apply to operate on the
    local domain. If false, it assumes that the generated code is already local)
    """

    def apply(self, ctx: Context, op: builtin.ModuleOp) -> None:
        if self.strategy not in self.STRATEGIES:
            raise ValueError(f"Unknown strategy: {self.strategy}")
        strategy = self.STRATEGIES[self.strategy](self.slices)

        rewrites: list[RewritePattern] = [
            AddHaloExchangeOps(strategy),
        ]

        if self.restrict_domain:
            rewrites.append(ChangeStoreOpSizes(strategy))

        PatternRewriteWalker(
            GreedyRewritePatternApplier(rewrites),
            apply_recursively=False,
        ).rewrite_module(op)


@dataclass(frozen=True)
class DmpToMpiPass(ModulePass):
    name = "dmp-to-mpi"

    mpi_init: bool = True

    generate_debug_prints: bool = False

    def apply(self, ctx: Context, op: builtin.ModuleOp) -> None:
        PatternRewriteWalker(
            GreedyRewritePatternApplier(
                [
                    LowerHaloExchangeToMpi(
                        self.mpi_init,
                        self.generate_debug_prints,
                    ),
                ]
            )
        ).rewrite_module(op)
        MpiLoopInvariantCodeMotion().rewrite_module(op)



#!/usr/bin/env python3
"""Utilities for profiling ASV benchmarks with a variety of tools."""

import cProfile
import gc
import subprocess
import time
from argparse import ArgumentParser, Namespace
from collections.abc import Callable
from pathlib import Path
from statistics import mean, median, stdev
from typing import Any, NamedTuple, cast

DEFAULT_OUTPUT_DIRECTORY = Path(__file__).parent / "profiles"
PROFILERS = ("run", "timeit", "snakeviz", "viztracer", "flameprof")


class Benchmark(NamedTuple):
    """A wrapper for a benchmark function with optional setup funtion."""

    body: Callable[[], Any]
    setup: Callable[[], Any] | None = None


def warmed_timeit(
    func: Callable[[], Any],
    setup: Callable[[], Any] | None = None,
    number: int = 100,
    warmup: int = 3,
) -> tuple[float, float, float]:
    """
    Time the contents of a class method with setup and warmup.

    Derived from <https://github.com/python/cpython/blob/3.13/Lib/timeit.py>.
    """

    class EmptyBenchmarkClass:
        """A class with an empty method call as a baseline for timing."""

        def empty(self) -> None:
            """An empty method call."""
            pass

    benchmark_class_empty = EmptyBenchmarkClass().empty

    # Inspired by timeit, we disable garbage collection for less noise in
    # measurements
    gcold = gc.isenabled()
    gc.disable()
    # Pre-populate the arrays to avoid costs of re-sizing them
    times = [0.0 for _ in range(number)]
    offset = [0.0 for _ in range(number)]

    # Run the interleaved setup and function calls to warm up
    for _ in range(warmup):
        if setup is not None:
            setup()
        func()

    for i in range(number):
        # Optionally run setup code (for example resetting mutable state) before
        # each measurement iteration
        if setup is not None:
            setup()
        # Calculate the base cost of method invocation and timing overhead, so
        # we can offset our final measurements by it
        offset_start = time.perf_counter()
        benchmark_class_empty()
        offset_end = time.perf_counter()
        offset[i] = offset_end - offset_start
        # Time the actual function we want to measure
        func_start = time.perf_counter()
        func()
        func_end = time.perf_counter()
        times[i] = func_end - func_start

    # Re-enable the garbage collector if it was initially on
    if gcold:
        gc.enable()

    # Return the mean, median, and standard deviations of the measured times.
    # The mean offset is subtracted from the median time for the best
    # approximation given the data we can record
    return (
        mean(times) - mean(offset),
        median(times) - mean(offset),
        stdev(times) + stdev(offset),
    )


def parse_arguments(benchmark_names: list[str]) -> ArgumentParser:
    """Parse the arguments for the profiler tool."""
    parser = ArgumentParser()

    parser.add_argument(
        "test",
        choices=benchmark_names + ["all"],
        help="the name of the benchmark to run, `all` to run all benchmarks",
    )
    parser.add_argument(
        "profiler",
        choices=PROFILERS,
        nargs="?",
        default=PROFILERS[0],
        help="the profiler to use",
    )
    parser.add_argument(
        "-o",
        "--output",
        type=Path,
        default=DEFAULT_OUTPUT_DIRECTORY,
        help="the directory into which to write out the profile files",
    )
    parser.add_argument(
        "-q", "--quiet", action="store_true", help="don't show the profiler's UI"
    )

    return parser


def get_benchmark_runs(
    args: Namespace,
    benchmarks: dict[str, Benchmark],
) -> list[tuple[str, Benchmark]]:
    """Get the benchmark to profile."""
    if args.test == "all":
        return list(benchmarks.items())

    name = args.test
    test = benchmarks[name]
    return [(name, test)]


def run_benchmark(
    args: Namespace,
    benchmarks: dict[str, Benchmark],
    warmup: bool = False,
) -> None:
    """Directly run a benchmark."""
    benchmark_runs = get_benchmark_runs(args, benchmarks)
    for name, (test, setup) in benchmark_runs:
        if warmup:
            if setup is not None:
                setup()
            test()
        if setup is not None:
            setup()
        start_time = time.time()
        test()
        print(f"Test {name} ran in: {time.time() - start_time:.5f}s")


def timeit_benchmark(
    args: Namespace,
    benchmarks: dict[str, Benchmark],
) -> None:
    """Use a custom function based on timeit to run a benchmark."""
    benchmark_runs = get_benchmark_runs(args, benchmarks)
    for name, (test, setup) in benchmark_runs:
        me, _, std = warmed_timeit(test, setup=setup)
        print(f"Test {name} ran in: {me:.3g} ± {std:.3g}s")


def cprofile_benchmark(
    args: Namespace,
    benchmarks: dict[str, Benchmark],
    warmup: bool = False,
) -> Path:
    """Use cProfile to profile a benchmark."""
    benchmark_runs = get_benchmark_runs(args, benchmarks)
    if len(benchmark_runs) != 1:
        raise ValueError("Cannot profile multiple benchmarks together")
    name, (test, setup) = benchmark_runs[0]
    output_prof = args.output / f"{name}.prof"
    if warmup:
        if setup is not None:
            setup()
        test()
    if setup is not None:
        setup()
    profiler = cProfile.Profile()
    profiler.enable()
    test()
    profiler.disable()
    profiler.dump_stats(str(output_prof))
    return output_prof


def viztracer_benchmark(
    args: Namespace,
    benchmarks: dict[str, Benchmark],
    warmup: bool = True,
) -> Path:
    """Use VizTracer to profile a benchmark."""
    from viztracer import VizTracer  # pyright: ignore[reportMissingTypeStubs]

    benchmark_runs = get_benchmark_runs(args, benchmarks)
    if len(benchmark_runs) != 1:
        raise ValueError("Cannot profile multiple benchmarks together")
    name, (test, setup) = benchmark_runs[0]
    output_prof = args.output / f"{name}.json"
    if warmup:
        if setup is not None:
            setup()
        test()
    if setup is not None:
        setup()
    with VizTracer(output_file=str(output_prof)):
        test()
    return output_prof


def show(
    args: Namespace,
    output_prof: Path,
    tool: str,
    options: tuple[str, ...] | None = None,
) -> None:
    """Show the profile using the specified tool."""
    if args.quiet:
        return
    if options is None:
        options = cast(tuple[str], ())
    command = ["uv", "run", tool, output_prof, *options]
    subprocess.run(command, check=True)  # noqa: S603


def profile(
    benchmarks: dict[str, Benchmark],
    argv: list[str] | None = None,
) -> None:
    """Run the selected profiler."""
    if not benchmarks:
        raise ValueError("At least one benchmark must be provided to profile!")

    args = parse_arguments(list(benchmarks.keys())).parse_args(args=argv)

    match args.profiler:
        case "run":
            run_benchmark(args, benchmarks)
        case "timeit":
            timeit_benchmark(args, benchmarks)
        case "snakeviz":
            output_prof = cprofile_benchmark(args, benchmarks)
            show(args, output_prof, tool="snakeviz")
        case "flameprof":
            output_prof = cprofile_benchmark(args, benchmarks)
            show(args, output_prof, tool="flameprof")
        case "viztracer":
            output_prof = viztracer_benchmark(args, benchmarks)
            show(args, output_prof, tool="vizviewer")
        case _:
            raise ValueError("Invalid command!")



"""
Test the definition and usage of traits and interfaces.
"""

from __future__ import annotations

from abc import ABC
from collections.abc import Sequence
from dataclasses import dataclass
from typing import Any

import pytest

from xdsl.dialects import test
from xdsl.dialects.builtin import (
    DYNAMIC_INDEX,
    AnyTensorTypeConstr,
    AnyUnrankedMemRefTypeConstr,
    AnyUnrankedTensorTypeConstr,
    IntegerAttr,
    IntegerType,
    MemRefType,
    NoneAttr,
    StringAttr,
    SymbolRefAttr,
    TensorType,
    UnrankedTensorType,
    i1,
    i32,
    i64,
)
from xdsl.ir import Attribute, Operation, OpTrait, OpTraits, SSAValue
from xdsl.irdl import (
    Block,
    IRDLOperation,
    Region,
    attr_def,
    irdl_op_definition,
    lazy_traits_def,
    operand_def,
    opt_attr_def,
    opt_region_def,
    prop_def,
    region_def,
    result_def,
    traits_def,
    var_operand_def,
    var_result_def,
)
from xdsl.traits import (
    AlwaysSpeculatable,
    ConditionallySpeculatable,
    HasAncestor,
    HasParent,
    OptionalSymbolOpInterface,
    RecursivelySpeculatable,
    SameOperandsAndResultType,
    SymbolOpInterface,
    SymbolTable,
    is_speculatable,
)
from xdsl.utils.exceptions import VerifyException
from xdsl.utils.test_value import TestSSAValue


@dataclass(frozen=True)
class LargerResultTrait(OpTrait):
    """Check that the only result has a larger bitwidth than the operand."""

    def verify(self, op: Operation) -> None:
        # This function is never called in this test
        raise NotImplementedError()


@dataclass(frozen=True)
class LargerOperandTrait(OpTrait):
    """Check that the only operand has a larger bitwidth than the result."""

    def verify(self, op: Operation):
        # These asserts should be exceptions in a non-testing environment.
        assert len(op.results) == 1
        assert len(op.operands) == 1
        assert isinstance(op.results[0].type, IntegerType)
        assert isinstance(op.operands[0].type, IntegerType)
        if op.results[0].type.width.data >= op.operands[0].type.width.data:
            raise VerifyException(
                "Operation has a result bitwidth greater "
                "or equal to the operand bitwidth."
            )


@dataclass(frozen=True)
class BitwidthSumLessThanTrait(OpTrait):
    """
    Check that the sum of the bitwidths of the
    operands and results is less than a given value.
    """

    parameters: int

    @property
    def max_sum(self):
        return self.parameters

    def verify(self, op: Operation):
        sum_bitwidth = 0
        for operand in op.operands:
            # This assert should be an exception in a non-testing environment.
            assert isinstance(operand.type, IntegerType)
            sum_bitwidth += operand.type.width.data
        for result in op.results:
            # This assert should be an exception in a non-testing environment.
            assert isinstance(result.type, IntegerType)
            sum_bitwidth += result.type.width.data

        if sum_bitwidth >= self.max_sum:
            raise VerifyException(
                f"Operation has a bitwidth sum greater or equal to {self.max_sum}."
            )


@irdl_op_definition
class TestOp(IRDLOperation):
    name = "test.test"
    traits = traits_def(LargerOperandTrait(), BitwidthSumLessThanTrait(64))

    ops = operand_def(IntegerType)
    res = result_def(IntegerType)


def test_has_trait_object():
    """
    Test the `has_trait` `Operation` method on a simple operation definition.
    """
    assert TestOp.has_trait(LargerOperandTrait)
    assert not TestOp.has_trait(LargerResultTrait)
    assert not TestOp.has_trait(BitwidthSumLessThanTrait(0))
    assert TestOp.has_trait(BitwidthSumLessThanTrait(64))


def test_get_traits_of_type():
    """
    Test the `get_traits_of_type` `Operation` method
    on a simple operation definition.
    """
    assert TestOp.get_traits_of_type(LargerOperandTrait) == [LargerOperandTrait()]
    assert TestOp.get_traits_of_type(LargerResultTrait) == []
    assert TestOp.get_traits_of_type(BitwidthSumLessThanTrait) == [
        BitwidthSumLessThanTrait(64)
    ]


def test_verifier():
    """
    Check that the traits verifier are correctly called.
    """
    operand64 = TestSSAValue(i64)
    operand32 = TestSSAValue(i32)
    operand1 = TestSSAValue(i1)
    op = TestOp.create(operands=[operand1], result_types=[i32])

    message = (
        "Operation has a result bitwidth greater or equal to the operand bitwidth."
    )
    with pytest.raises(VerifyException, match=message):
        op.verify()

    op = TestOp.create(operands=[operand64], result_types=[i32])
    with pytest.raises(
        VerifyException, match="Operation has a bitwidth sum greater or equal to 64."
    ):
        op.verify()

    op = TestOp.create(operands=[operand32], result_types=[i1])
    op.verify()


def test_verifier_order():
    """
    Check that trait verifiers are called after IRDL verifiers.
    """
    op = TestOp.create(operands=[], result_types=[i1])
    with pytest.raises(VerifyException, match="Expected 1 operand, but got 0"):
        op.verify()


class LargerOperandOp(IRDLOperation, ABC):
    traits = traits_def(LargerOperandTrait())


@irdl_op_definition
class TestCopyOp(LargerOperandOp):
    name = "test.test_copy"

    traits = OpTraits(LargerOperandOp.traits.traits | {BitwidthSumLessThanTrait(64)})


def test_trait_inheritance():
    """
    Check that traits are correctly inherited from parent classes.
    """
    assert TestCopyOp.traits == traits_def(
        LargerOperandTrait(), BitwidthSumLessThanTrait(64)
    )


@irdl_op_definition
class NoTraitsOp(IRDLOperation):
    name = "test.no_traits_op"


def test_traits_undefined():
    """Check that traits are defaulted to the empty set."""
    assert NoTraitsOp.traits == traits_def()


class WrongTraitsType(IRDLOperation):
    name = "test.no_traits"

    traits = 1  # pyright: ignore[reportAssignmentType]


def test_traits_wrong_type():
    with pytest.raises(Exception):
        irdl_op_definition(WrongTraitsType)


class GetNumResultsTrait(OpTrait):
    """
    An example of an MLIR interface, using traits.
    It provides a method to get the number of results of an operation.
    """

    # This is the default implementation of a trait (interface) method.
    @staticmethod
    def get_num_results(op: Operation):
        """
        Get the number of results of the operation
        """
        return len(op.results)


class GetNumResultsTraitForOpWithOneResult(GetNumResultsTrait):
    """
    Instance of the trait for the case where an operation has only a single result.
    """

    @staticmethod
    def get_num_results(op: Operation):
        return 1


@irdl_op_definition
class HasInterfaceOp(IRDLOperation):
    name = "test.op_with_interface"
    traits = traits_def(GetNumResultsTraitForOpWithOneResult())

    res = result_def(IntegerType)


def test_interface():
    """
    Test the features of a trait with methods (An MLIR interface).
    """
    op = HasInterfaceOp.create(result_types=(i32,))
    trait = HasInterfaceOp.get_trait(GetNumResultsTrait)
    assert trait is not None
    assert 1 == trait.get_num_results(op)


def test_get_trait_specialized():
    """
    Test get_trait and has_trait in the case where the trait is a child class of the
    trait we want.
    """
    assert HasInterfaceOp.has_trait(GetNumResultsTrait)
    assert HasInterfaceOp.has_trait(GetNumResultsTraitForOpWithOneResult)
    assert (
        HasInterfaceOp.get_trait(GetNumResultsTrait)
        == GetNumResultsTraitForOpWithOneResult()
    )
    assert HasInterfaceOp.get_traits_of_type(GetNumResultsTrait) == [
        GetNumResultsTraitForOpWithOneResult()
    ]


def test_symbol_op_interface():
    """
    Test that operations that conform to SymbolOpInterface have necessary attributes.
    """

    @irdl_op_definition
    class NoSymNameOp(IRDLOperation):
        name = "no_sym_name"
        traits = traits_def(SymbolOpInterface())

    op0 = NoSymNameOp()

    with pytest.raises(
        VerifyException, match='Operation no_sym_name must have a "sym_name" attribute'
    ):
        op0.verify()

    @irdl_op_definition
    class SymNameWrongTypeOp(IRDLOperation):
        name = "wrong_sym_name_type"

        sym_name = attr_def(IntegerAttr)
        traits = traits_def(SymbolOpInterface())

    op1 = SymNameWrongTypeOp(
        attributes={"sym_name": IntegerAttr.from_int_and_width(1, 32)}
    )

    with pytest.raises(
        VerifyException,
        match='Operation wrong_sym_name_type must have a "sym_name" attribute',
    ):
        op1.verify()

    @irdl_op_definition
    class SymNameOp(IRDLOperation):
        name = "sym_name"

        sym_name = attr_def(StringAttr)
        traits = traits_def(SymbolOpInterface())

    op2 = SymNameOp(attributes={"sym_name": StringAttr("symbol_name")})
    op2.verify()


def test_optional_symbol_op_interface():
    """
    Test that operations that conform to OptionalSymbolOpInterface have the necessary attributes.
    """

    @irdl_op_definition
    class OptionalSymNameOp(IRDLOperation):
        name = "no_sym_name"

        sym_name = opt_attr_def(StringAttr)

        traits = traits_def(OptionalSymbolOpInterface())

    no_symbol = OptionalSymNameOp()
    interface = no_symbol.get_trait(SymbolOpInterface)
    assert interface is not None
    assert interface.is_optional_symbol(no_symbol)
    no_symbol.verify()
    assert interface.get_sym_attr_name(no_symbol) is None

    symbol = OptionalSymNameOp(attributes={"sym_name": StringAttr("main")})
    interface = symbol.get_trait(SymbolOpInterface)
    assert interface is not None
    assert interface.is_optional_symbol(symbol)
    symbol.verify()
    assert interface.get_sym_attr_name(symbol) == StringAttr("main")


@irdl_op_definition
class SymbolOp(IRDLOperation):
    name = "test.symbol"

    sym_name = attr_def(StringAttr)

    traits = traits_def(SymbolOpInterface())

    def __init__(self, name: str):
        return super().__init__(attributes={"sym_name": StringAttr(name)})


@irdl_op_definition
class PropSymbolOp(IRDLOperation):
    name = "test.symbol"

    sym_name = prop_def(StringAttr)

    traits = traits_def(SymbolOpInterface())

    def __init__(self, name: str):
        return super().__init__(properties={"sym_name": StringAttr(name)})


@pytest.mark.parametrize("SymbolOp", [SymbolOp, PropSymbolOp])
def test_symbol_table(SymbolOp: type[PropSymbolOp | SymbolOp]):
    # Some helper classes
    @irdl_op_definition
    class SymbolTableOp(IRDLOperation):
        name = "test.symbol_table"

        sym_name = opt_attr_def(StringAttr)

        one = region_def()
        two = opt_region_def()

        traits = traits_def(SymbolTable(), OptionalSymbolOpInterface())

    # Check that having a single region is verified
    op = SymbolTableOp(regions=[Region(), Region()])
    with pytest.raises(
        VerifyException,
        match="Operations with a 'SymbolTable' must have exactly one region",
    ):
        op.verify()

    # Check that having a single block is verified
    blocks = [Block(), Block()]
    op = SymbolTableOp(regions=[Region(blocks), []])
    with pytest.raises(
        VerifyException,
        match="Operations with a 'SymbolTable' must have exactly one block",
    ):
        op.verify()

    terminator = test.TestTermOp()

    # Check that symbol uniqueness is verified
    symbol = SymbolOp("name")
    dup_symbol = SymbolOp("name")
    op = SymbolTableOp(
        regions=[Region(Block([symbol, dup_symbol, terminator.clone()])), []]
    )
    with pytest.raises(
        VerifyException,
        match='Redefinition of symbol "name"',
    ):
        op.verify()

    # Check a flat happy case, with symbol lookup
    symbol = SymbolOp("name")

    op = SymbolTableOp(regions=[Region(Block([symbol, terminator.clone()])), []])
    op.verify()

    assert SymbolTable.lookup_symbol(op, "name") is symbol
    assert SymbolTable.lookup_symbol(op, "that_other_name") is None

    # Check a nested happy case, with symbol lookup
    symbol = SymbolOp("name")

    nested = SymbolTableOp(
        regions=[Region(Block([symbol, terminator.clone()])), []],
        attributes={"sym_name": StringAttr("nested")},
    )
    op = SymbolTableOp(regions=[Region(Block([nested, terminator.clone()])), []])
    op.verify()

    assert SymbolTable.lookup_symbol(op, "name") is None
    assert SymbolTable.lookup_symbol(op, SymbolRefAttr("nested", ["name"])) is symbol


@irdl_op_definition
class HasLazyParentOp(IRDLOperation):
    """An operation with traits that are defined "lazily"."""

    name = "test.has_lazy_parent"

    traits = lazy_traits_def(lambda: (HasParent(TestOp),))


def test_lazy_parent():
    """Test the trait infrastructure for an operation that defines a trait "lazily"."""
    op = HasLazyParentOp.create()
    assert len(op.get_traits_of_type(HasParent)) != 0
    assert op.get_traits_of_type(HasParent)[0].op_types == (TestOp,)
    assert op.has_trait(HasParent(TestOp))
    assert op.traits == traits_def(HasParent(TestOp))


@irdl_op_definition
class AncestorOp(IRDLOperation):
    name = "test.ancestor"

    traits = traits_def(HasAncestor(TestOp))


def test_has_ancestor():
    op = AncestorOp()

    assert op.get_traits_of_type(HasAncestor) == [HasAncestor(TestOp)]
    assert op.has_trait(HasAncestor(TestOp))

    with pytest.raises(
        VerifyException, match="'test.ancestor' expects ancestor op 'test.test'"
    ):
        op.verify()


def test_insert_or_update():
    @irdl_op_definition
    class SymbolTableOp(IRDLOperation):
        name = "test.symbol_table"

        reg = region_def()

        traits = traits_def(SymbolTable())

    # Check a flat happy case, with symbol lookup
    symbol = SymbolOp("name")
    symbol2 = SymbolOp("name2")
    terminator = test.TestTermOp()

    op = SymbolTableOp(regions=[Region(Block([symbol, terminator]))])
    op.verify()

    trait = op.get_trait(SymbolTable)
    assert trait is not None

    assert trait.insert_or_update(op, symbol.clone()) is symbol
    assert len(op.reg.ops) == 2

    assert trait.insert_or_update(op, symbol2) is None
    assert len(op.reg.ops) == 3
    assert symbol2 in list(op.reg.ops)


def nonpure():
    return TestOp.create()


def pure():
    return test.TestPureOp.create()


@pytest.mark.parametrize(
    ("trait", "speculatability", "nested_ops"),
    [
        ([], False, []),
        ([], False, [nonpure()]),
        ([], False, [pure()]),
        ([AlwaysSpeculatable()], True, []),
        ([AlwaysSpeculatable()], True, [nonpure()]),
        ([AlwaysSpeculatable()], True, [pure()]),
        ([RecursivelySpeculatable()], True, []),
        ([RecursivelySpeculatable()], False, [nonpure()]),
        ([RecursivelySpeculatable()], True, [pure()]),
    ],
)
def test_speculability(
    trait: tuple[ConditionallySpeculatable] | tuple[()],
    speculatability: bool,
    nested_ops: Sequence[Operation],
):
    @irdl_op_definition
    class SupeculatabilityTestOp(IRDLOperation):
        name = "test.speculatability"
        region = region_def()

        traits = traits_def(*trait)

    op = SupeculatabilityTestOp(regions=[Region(Block(nested_ops))])
    optrait = op.get_trait(ConditionallySpeculatable)

    if trait:
        assert optrait is not None
        assert optrait.is_speculatable(op) is speculatability
    else:
        assert optrait is None

    assert is_speculatable(op) is speculatability


@pytest.mark.parametrize(
    ("operands", "result_types"),
    [
        ([()], [()]),
        ([()], (test.TestType("foo"),)),
        ((TestSSAValue(test.TestType("foo")),), [()]),
    ],
)
def test_same_operands_and_result_type_trait_for_scalar_types(
    operands: tuple[SSAValue] | tuple[()],
    result_types: tuple[test.TestType] | tuple[()],
):
    @irdl_op_definition
    class SameOperandsAndResultTypeOp(IRDLOperation):
        name = "test.same_operand_and_result_type"

        ops = var_operand_def(test.TestType("foo"))
        res = var_result_def(test.TestType("foo"))

        traits = traits_def(SameOperandsAndResultType())

    op = SameOperandsAndResultTypeOp(operands=operands, result_types=result_types)

    with pytest.raises(
        VerifyException, match="requires at least one result or operand"
    ):
        op.verify()


@irdl_op_definition
class SameOperandsAndResultTypeOp(IRDLOperation):
    name = "test.same_operand_and_result_type"

    ops = var_operand_def(
        MemRefType.constr()
        | AnyUnrankedMemRefTypeConstr
        | AnyUnrankedTensorTypeConstr
        | AnyTensorTypeConstr
    )

    res = var_result_def(
        MemRefType.constr()
        | AnyUnrankedMemRefTypeConstr
        | AnyUnrankedTensorTypeConstr
        | AnyTensorTypeConstr
    )

    traits = traits_def(SameOperandsAndResultType())


@pytest.mark.parametrize(
    (
        "operand1_and_result_element_type",
        "operand_and_result_shape1",
        "result_element_type2",
        "result_shape2",
    ),
    [
        (
            test.TestType("foo"),
            [2, 3],
            test.TestType("qux"),
            [2, 3],
        ),
        (
            test.TestType("foo"),
            [2, 3],
            test.TestType("foo"),
            [2, 4],
        ),
        (
            test.TestType("qux"),
            [2, 3],
            test.TestType("foo"),
            [2, 3],
        ),
        (
            test.TestType("foo"),
            [2, 4],
            test.TestType("foo"),
            [2, 3],
        ),
    ],
)
def test_same_operands_and_result_type_trait_for_result_element_type_of_shaped_types(
    operand1_and_result_element_type: Attribute,
    operand_and_result_shape1: tuple[int],
    result_element_type2: Attribute,
    result_shape2: tuple[int],
):
    op = SameOperandsAndResultTypeOp(
        operands=[
            TestSSAValue(
                TensorType(operand1_and_result_element_type, operand_and_result_shape1)
            )
        ],
        result_types=[
            [
                TensorType(operand1_and_result_element_type, operand_and_result_shape1),
                TensorType(result_element_type2, result_shape2),
            ],
        ],
    )

    with pytest.raises(
        VerifyException,
        match="requires the same type for all operands and results",
    ):
        op.verify()


@pytest.mark.parametrize(
    "operands_num",
    [1, 2, 3],
)
@pytest.mark.parametrize(
    "results_num",
    [1, 2, 3],
)
@pytest.mark.parametrize(
    (
        "operand_type",
        "result_type",
    ),
    [
        (
            TensorType(test.TestType("foo"), [2, 3]),
            TensorType(test.TestType("qux"), [2, 3]),
        ),
        (
            TensorType(test.TestType("foo"), [2, 3]),
            TensorType(test.TestType("foo"), [2, 4]),
        ),
        (
            TensorType(test.TestType("qux"), [2, 3]),
            TensorType(test.TestType("foo"), [2, 3]),
        ),
        (
            TensorType(test.TestType("foo"), [2, 4]),
            TensorType(test.TestType("foo"), [2, 3]),
        ),
        (
            MemRefType(test.TestType("foo"), [2, 3]),
            MemRefType(test.TestType("qux"), [2, 3]),
        ),
        (
            MemRefType(test.TestType("foo"), [2, 3]),
            MemRefType(test.TestType("foo"), [2, 4]),
        ),
        (
            MemRefType(test.TestType("qux"), [2, 3]),
            MemRefType(test.TestType("foo"), [2, 3]),
        ),
        (
            MemRefType(test.TestType("foo"), [2, 4]),
            MemRefType(test.TestType("foo"), [2, 3]),
        ),
    ],
)
def test_same_operands_and_result_type_trait_for_element_type_of_shaped_types(
    operand_type: TensorType[Any],
    result_type: TensorType[Any],
    operands_num: int,
    results_num: int,
):
    op = SameOperandsAndResultTypeOp(
        operands=[
            [
                TestSSAValue(operand_type),
            ]
            * operands_num,
        ],
        result_types=[[result_type] * results_num],
    )

    with pytest.raises(
        VerifyException,
        match="requires the same type for all operands and results",
    ):
        op.verify()


@pytest.mark.parametrize(
    (
        "element_type",
        "shape",
        "operand1_and_result_encoding",
        "result_encoding2",
    ),
    [
        (
            test.TestType("foo"),
            [2, 3],
            StringAttr("bar"),
            StringAttr("baz"),
        ),
        (
            test.TestType("foo"),
            [2, 3],
            StringAttr("baz"),
            StringAttr("bar"),
        ),
        (
            test.TestType("foo"),
            [2, 3],
            StringAttr("bar"),
            NoneAttr(),
        ),
        (
            test.TestType("foo"),
            [2, 3],
            NoneAttr(),
            StringAttr("bar"),
        ),
    ],
)
def test_same_operands_and_result_type_trait_for_result_encoding_of_shaped_types(
    element_type: Attribute,
    shape: tuple[int],
    operand1_and_result_encoding: Attribute,
    result_encoding2: Attribute,
):
    op = SameOperandsAndResultTypeOp(
        operands=[
            [
                TestSSAValue(
                    TensorType(
                        element_type,
                        shape,
                        operand1_and_result_encoding,
                    )
                ),
            ],
        ],
        result_types=[
            [
                TensorType(element_type, shape, operand1_and_result_encoding),
                TensorType(element_type, shape, result_encoding2),
            ]
        ],
    )

    with pytest.raises(
        VerifyException,
        match="requires the same encoding for all operands and results",
    ):
        op.verify()


@pytest.mark.parametrize(
    "operands_num",
    [1, 2, 3],
)
@pytest.mark.parametrize(
    "results_num",
    [1, 2, 3],
)
@pytest.mark.parametrize(
    (
        "element_type",
        "shape",
        "operand_encoding",
        "result_encoding",
    ),
    [
        (
            test.TestType("foo"),
            [2, 3],
            StringAttr("bar"),
            StringAttr("baz"),
        ),
        (
            test.TestType("foo"),
            [2, 3],
            StringAttr("baz"),
            StringAttr("bar"),
        ),
        (
            test.TestType("foo"),
            [2, 3],
            StringAttr("bar"),
            NoneAttr(),
        ),
        (
            test.TestType("foo"),
            [2, 3],
            NoneAttr(),
            StringAttr("bar"),
        ),
    ],
)
def test_same_operands_and_result_type_trait_for_encoding_of_shaped_types(
    element_type: Attribute,
    shape: tuple[int],
    operand_encoding: Attribute,
    result_encoding: Attribute,
    operands_num: int,
    results_num: int,
):
    op = SameOperandsAndResultTypeOp(
        operands=[
            [
                TestSSAValue(
                    TensorType(
                        element_type,
                        shape,
                        operand_encoding,
                    )
                ),
            ]
            * operands_num,
        ],
        result_types=[[TensorType(element_type, shape, result_encoding)] * results_num],
    )

    with pytest.raises(
        VerifyException,
        match="requires the same encoding for all operands and results",
    ):
        op.verify()


@pytest.mark.parametrize(
    "operands_num",
    [1, 2, 3],
)
@pytest.mark.parametrize(
    "results_num",
    [1, 2, 3],
)
@pytest.mark.parametrize(
    (
        "operand1_and_result_shape",
        "operand2_shape",
    ),
    [
        (
            [1],
            [1],
        ),
        (
            [2, 3],
            [2, 3],
        ),
        (
            [2, 3],
            [2, DYNAMIC_INDEX],
        ),
        (
            [2, 4],
            [2, DYNAMIC_INDEX],
        ),
        (
            [2, DYNAMIC_INDEX],
            [2, DYNAMIC_INDEX],
        ),
    ],
)
def test_same_operands_and_result_type_trait_for_ranked_mixed_shapes(
    operand1_and_result_shape: tuple[int],
    operand2_shape: tuple[int],
    operands_num: int,
    results_num: int,
):
    op = SameOperandsAndResultTypeOp(
        operands=[
            [
                TestSSAValue(
                    TensorType(test.TestType("foo"), operand1_and_result_shape)
                ),
                TestSSAValue(TensorType(test.TestType("foo"), operand2_shape)),
            ]
            * operands_num,
        ],
        result_types=[
            [TensorType(test.TestType("foo"), operand1_and_result_shape)] * results_num
        ],
    )

    op.verify()


@pytest.mark.parametrize(
    "operands_num",
    [1, 2, 3],
)
@pytest.mark.parametrize(
    "results_num",
    [1, 2, 3],
)
@pytest.mark.parametrize(
    ("operand1_and_result_shape",),
    [
        ([1],),
        ([2, 3],),
        ([2, DYNAMIC_INDEX],),
        ([DYNAMIC_INDEX, DYNAMIC_INDEX],),
    ],
)
def test_same_operands_and_result_type_trait_for_mixed_rank_and_mixed_shapes(
    operand1_and_result_shape: tuple[int],
    operands_num: int,
    results_num: int,
):
    op = SameOperandsAndResultTypeOp(
        operands=[
            [
                TestSSAValue(
                    TensorType(test.TestType("foo"), operand1_and_result_shape)
                ),
                TestSSAValue(UnrankedTensorType(test.TestType("foo"))),
            ]
            * operands_num,
        ],
        result_types=[
            [TensorType(test.TestType("foo"), operand1_and_result_shape)] * results_num
        ],
    )

    op.verify()


@irdl_op_definition
class TestModifyTraitsOp(IRDLOperation):
    name = "test.test_modify_traits"


class AlwaysFailsTrait(OpTrait):
    def verify(self, op: Operation) -> None:
        raise VerifyException("Nope")


def test_modify_traits():
    op = TestModifyTraitsOp()

    op.verify()

    TestModifyTraitsOp.traits.add_trait(AlwaysFailsTrait())

    with pytest.raises(VerifyException, match="Nope"):
        op.verify()



"""
Test the usage of scf dialect.
"""

from typing import cast

import pytest

from xdsl.builder import Builder
from xdsl.dialects.arith import ConstantOp
from xdsl.dialects.builtin import IndexType, ModuleOp, i32, i64
from xdsl.dialects.scf import (
    ForOp,
    IfOp,
    ParallelOp,
    ReduceOp,
    ReduceReturnOp,
    WhileOp,
    YieldOp,
)
from xdsl.dialects.test import TestTermOp
from xdsl.ir import Block, BlockArgument, Region
from xdsl.utils.exceptions import DiagnosticException, VerifyException


def test_for_with_loop_carried_verify():
    """Test for with loop-carried variables"""

    lower = ConstantOp.from_int_and_width(0, IndexType())
    upper = ConstantOp.from_int_and_width(42, IndexType())
    step = ConstantOp.from_int_and_width(3, IndexType())
    carried = ConstantOp.from_int_and_width(1, IndexType())

    @Builder.implicit_region((IndexType(), IndexType()))
    def body(_: tuple[BlockArgument, ...]) -> None:
        YieldOp(carried)

    for_op = ForOp(lower, upper, step, [carried], body)

    assert for_op.lb is lower.result
    assert for_op.ub is upper.result
    assert for_op.step is step.result
    assert for_op.iter_args == tuple([carried.result])
    assert for_op.body is body

    assert len(for_op.results) == 1
    assert for_op.results[0].type == carried.result.type
    assert tuple(for_op.operands) == (
        lower.result,
        upper.result,
        step.result,
        carried.result,
    )
    assert for_op.regions == (body,)
    assert for_op.attributes == {}

    for_op.verify()


def test_for_without_loop_carried_verify():
    """Test for without loop-carried variables"""

    lower = ConstantOp.from_int_and_width(0, IndexType())
    upper = ConstantOp.from_int_and_width(42, IndexType())
    step = ConstantOp.from_int_and_width(3, IndexType())

    @Builder.implicit_region((IndexType(),))
    def body(_: tuple[BlockArgument, ...]) -> None:
        YieldOp()

    for_op = ForOp(lower, upper, step, [], body)

    assert for_op.lb is lower.result
    assert for_op.ub is upper.result
    assert for_op.step is step.result
    assert for_op.iter_args == tuple()
    assert for_op.body is body

    assert len(for_op.results) == 0
    assert tuple(for_op.operands) == (
        lower.result,
        upper.result,
        step.result,
    )
    assert for_op.regions == (body,)
    assert for_op.attributes == {}

    for_op.verify()


def test_parallel_no_init_vals():
    lbi = ConstantOp.from_int_and_width(0, IndexType())
    lbj = ConstantOp.from_int_and_width(1, IndexType())
    lbk = ConstantOp.from_int_and_width(18, IndexType())

    ubi = ConstantOp.from_int_and_width(10, IndexType())
    ubj = ConstantOp.from_int_and_width(110, IndexType())
    ubk = ConstantOp.from_int_and_width(92, IndexType())

    si = ConstantOp.from_int_and_width(1, IndexType())
    sj = ConstantOp.from_int_and_width(3, IndexType())
    sk = ConstantOp.from_int_and_width(8, IndexType())

    body = Region()

    lowerBounds = [lbi, lbj, lbk]
    upperBounds = [ubi, ubj, ubk]
    steps = [si, sj, sk]

    p = ParallelOp(lowerBounds, upperBounds, steps, body)

    assert isinstance(p, ParallelOp)
    assert p.lowerBound == tuple(l.result for l in lowerBounds)
    assert p.upperBound == tuple(l.result for l in upperBounds)
    assert p.step == tuple(l.result for l in steps)
    assert p.body is body


def test_parallel_with_init_vals():
    lbi = ConstantOp.from_int_and_width(0, IndexType())
    ubi = ConstantOp.from_int_and_width(10, IndexType())
    si = ConstantOp.from_int_and_width(1, IndexType())

    body = Region()

    init_val = ConstantOp.from_int_and_width(10, i32)

    initVals = [init_val]

    lowerBounds = [lbi]
    upperBounds = [ubi]
    steps = [si]

    p = ParallelOp(lowerBounds, upperBounds, steps, body, initVals)

    assert isinstance(p, ParallelOp)
    assert p.lowerBound == tuple(l.result for l in lowerBounds)
    assert p.upperBound == tuple(l.result for l in upperBounds)
    assert p.step == tuple(l.result for l in steps)
    assert p.body is body
    assert p.initVals == tuple(l.result for l in initVals)


def test_parallel_verify_one_block():
    lbi = ConstantOp.from_int_and_width(0, IndexType())
    ubi = ConstantOp.from_int_and_width(10, IndexType())
    si = ConstantOp.from_int_and_width(1, IndexType())

    body = Region()
    p = ParallelOp([lbi], [ubi], [si], body)
    with pytest.raises(DiagnosticException):
        p.verify()


def test_parallel_verify_num_bounds_equal():
    lbi = ConstantOp.from_int_and_width(0, IndexType())
    ubi = ConstantOp.from_int_and_width(10, IndexType())
    si = ConstantOp.from_int_and_width(1, IndexType())

    lowerBounds = [lbi, lbi]
    upperBounds = [ubi]
    steps = [si]

    body = Region(Block())

    p = ParallelOp(lowerBounds, upperBounds, steps, body)
    with pytest.raises(VerifyException):
        p.verify()

    upperBounds = [ubi, ubi]

    body2 = Region(Block())
    p2 = ParallelOp(lowerBounds, upperBounds, steps, body2)
    with pytest.raises(VerifyException):
        p2.verify()

    lowerBounds = [lbi]
    upperBounds = [ubi]
    steps = [si, si]

    body3 = Region(Block())
    p3 = ParallelOp(lowerBounds, upperBounds, steps, body3)
    with pytest.raises(VerifyException):
        p3.verify()


def test_parallel_verify_only_induction_in_block():
    lbi = ConstantOp.from_int_and_width(0, IndexType())
    ubi = ConstantOp.from_int_and_width(10, IndexType())
    si = ConstantOp.from_int_and_width(1, IndexType())

    init_val = ConstantOp.from_int_and_width(10, i32)

    initVals = [init_val]

    b = Block(arg_types=[IndexType(), i32])
    b.add_op(YieldOp(init_val))

    body = Region(b)
    p = ParallelOp([lbi], [ubi], [si], body, initVals)
    with pytest.raises(VerifyException):
        p.verify()

    b2 = Block(arg_types=[IndexType(), i32, i32])
    b2.add_op(YieldOp(init_val))
    body2 = Region(b2)
    p2 = ParallelOp([lbi], [ubi], [si], body2, initVals)
    with pytest.raises(VerifyException):
        p2.verify()


def test_parallel_block_arg_indextype():
    lbi = ConstantOp.from_int_and_width(0, IndexType())
    ubi = ConstantOp.from_int_and_width(10, IndexType())
    si = ConstantOp.from_int_and_width(1, IndexType())

    b = Block(arg_types=[IndexType()])
    b.add_op(ReduceOp())

    body = Region(b)
    p = ParallelOp([lbi], [ubi], [si], body)
    p.verify()

    b2 = Block(arg_types=[i32])
    b2.add_op(ReduceOp())
    body2 = Region(b2)
    p2 = ParallelOp([lbi], [ubi], [si], body2)
    with pytest.raises(VerifyException):
        p2.verify()


def test_parallel_verify_reduction_and_block_type():
    lbi = ConstantOp.from_int_and_width(0, IndexType())
    ubi = ConstantOp.from_int_and_width(10, IndexType())
    si = ConstantOp.from_int_and_width(1, IndexType())

    init_val = ConstantOp.from_int_and_width(10, i32)

    initVals = [init_val]

    b = Block(arg_types=[IndexType()])

    reduce_constant = ConstantOp.from_int_and_width(100, i32)
    rro = ReduceReturnOp(reduce_constant)
    reduce_block = Block(arg_types=[i32, i32])
    reduce_block.add_ops([reduce_constant, rro])

    b.add_op(ReduceOp((init_val,), (Region(reduce_block),)))

    body = Region(b)
    p = ParallelOp([lbi], [ubi], [si], body, initVals)
    # This should verify
    p.verify()


def test_parallel_verify_reduction_and_block_type_fails():
    lbi = ConstantOp.from_int_and_width(0, IndexType())
    ubi = ConstantOp.from_int_and_width(10, IndexType())
    si = ConstantOp.from_int_and_width(1, IndexType())

    init_val = ConstantOp.from_int_and_width(10, i64)

    initVals = [init_val]

    b = Block(arg_types=[IndexType()])

    reduce_constant = ConstantOp.from_int_and_width(100, i32)
    rro = ReduceReturnOp(reduce_constant)
    reduce_block = Block(arg_types=[i32, i32])
    reduce_block.add_ops([reduce_constant, rro])

    b.add_op(ReduceOp((init_val,), (Region(reduce_block),)))

    body = Region(b)
    p = ParallelOp([lbi], [ubi], [si], body, initVals)
    with pytest.raises(VerifyException):
        p.verify()


def test_reduce_op():
    init_val = ConstantOp.from_int_and_width(10, i32)

    reduce_op = ReduceOp((init_val,), (Region(Block(arg_types=[i32, i32])),))

    assert reduce_op.args[0] is init_val.results[0]
    assert reduce_op.args[0].type is i32
    assert len(reduce_op.reductions[0].blocks) == 1
    assert len(reduce_op.reductions[0].block.args) == 2
    assert (
        reduce_op.reductions[0].block.args[0].type
        == reduce_op.reductions[0].block.args[0].type
        == i32
    )


def test_reduce_op_num_block_args():
    init_val = ConstantOp.from_int_and_width(10, i32)
    reduce_constant = ConstantOp.from_int_and_width(100, i32)

    with pytest.raises(
        VerifyException,
        match="scf.reduce block must have exactly two arguments, but ",
    ):
        rro = ReduceReturnOp(reduce_constant)
        ReduceOp(
            (init_val,), (Region(Block([rro], arg_types=[i32, i32, i32])),)
        ).verify()

    with pytest.raises(
        VerifyException,
        match="scf.reduce block must have exactly two arguments, but ",
    ):
        rro = ReduceReturnOp(reduce_constant)
        ReduceOp((init_val,), (Region(Block([rro], arg_types=[i32])),)).verify()

    with pytest.raises(
        VerifyException,
        match="scf.reduce block must have exactly two arguments, but ",
    ):
        rro = ReduceReturnOp(reduce_constant)
        ReduceOp((init_val,), (Region(Block([rro], arg_types=[])),)).verify()


def test_reduce_op_num_block_arg_types():
    init_val = ConstantOp.from_int_and_width(10, i32)
    reduce_constant = ConstantOp.from_int_and_width(100, i32)

    with pytest.raises(
        VerifyException,
        match="scf.reduce block argument types must be the same but have",
    ):
        rro = ReduceReturnOp(reduce_constant)
        ReduceOp((init_val,), (Region(Block([rro], arg_types=[i32, i64])),)).verify()

    with pytest.raises(
        VerifyException,
        match="scf.reduce block argument types must be the same but have",
    ):
        rro = ReduceReturnOp(reduce_constant)
        ReduceOp((init_val,), (Region(Block([rro], arg_types=[i64, i32])),)).verify()


def test_reduce_op_num_block_arg_types_match_operand_type():
    init_val = ConstantOp.from_int_and_width(10, i32)

    with pytest.raises(VerifyException):
        ReduceOp((init_val,), (Region(Block(arg_types=[i64, i64])),)).verify()


def test_reduce_return_op_at_end():
    reduce_constant = ConstantOp.from_int_and_width(100, i32)
    rro = ReduceReturnOp(reduce_constant)
    reduce_block = Block(arg_types=[i32, i32])
    reduce_block.add_ops([reduce_constant, rro])

    init_val = ConstantOp.from_int_and_width(10, i32)
    ReduceOp((init_val,), (Region(reduce_block),)).verify()

    with pytest.raises(
        VerifyException,
        match="'scf.reduce' terminates with operation test.termop instead of scf.reduce.return",
    ):
        ReduceOp(
            (init_val,), (Region(Block([TestTermOp.create()], arg_types=[i32, i32])),)
        ).verify()


def test_reduce_return_type_is_arg_type():
    reduce_constant = ConstantOp.from_int_and_width(100, i32)
    rro = ReduceReturnOp(reduce_constant)
    reduce_block = Block(arg_types=[i32, i32])
    reduce_block.add_ops([reduce_constant, rro])

    init_val = ConstantOp.from_int_and_width(10, i64)
    with pytest.raises(VerifyException):
        ReduceOp((init_val,), (Region(reduce_block),)).verify()


def test_reduce_return_op():
    reduce_constant = ConstantOp.from_int_and_width(100, i32)
    rro = ReduceReturnOp(reduce_constant)

    assert rro.result is reduce_constant.results[0]
    assert rro.result.type is i32


def test_reduce_return_type_is_operand_type():
    reduce_constant = ConstantOp.from_int_and_width(100, i32)
    reduce_constant_wrong_type = ConstantOp.from_int_and_width(100, i64)
    rro = ReduceReturnOp(reduce_constant_wrong_type)
    reduce_block = Block(arg_types=[i32, i32])
    reduce_block.add_ops([reduce_constant, rro])

    init_val = ConstantOp.from_int_and_width(10, i32)
    with pytest.raises(
        VerifyException,
        match="scf.reduce.return result type at end of scf.reduce block must",
    ):
        ReduceOp((init_val,), (Region(reduce_block),)).verify()


def test_empty_else():
    # create if without an else block:
    m = ModuleOp(
        [
            t := ConstantOp.from_int_and_width(1, 1),
            IfOp(
                t,
                [],
                [
                    YieldOp(),
                ],
            ),
        ]
    )

    assert len(cast(IfOp, list(m.ops)[1]).false_region.blocks) == 0


def test_while():
    before_block = Block(arg_types=(i32, i32))
    after_block = Block(arg_types=(i32, i32))
    a = ConstantOp.from_int_and_width(0, i32)
    b = ConstantOp.from_int_and_width(0, i32)
    while_loop = WhileOp(
        arguments=[a, b],
        result_types=[i32, i32],
        before_region=[before_block],
        after_region=[after_block],
    )
    assert (len(while_loop.results)) == 2
    assert (len(while_loop.operands)) == 2



"""
Test the definition of attributes and their constraints.
"""

from __future__ import annotations

from collections.abc import Sequence
from dataclasses import dataclass
from enum import auto
from io import StringIO
from typing import Annotated, Any, Generic, TypeAlias, TypeVar, cast

import pytest

from xdsl.dialects.builtin import (
    IndexType,
    IntAttr,
    IntegerAttr,
    IntegerType,
    NoneAttr,
    Signedness,
)
from xdsl.ir import (
    Attribute,
    BitEnumAttribute,
    Data,
    EnumAttribute,
    ParametrizedAttribute,
    SpacedOpaqueSyntaxAttribute,
    StrEnum,
    TypedAttribute,
)
from xdsl.irdl import (
    AnyAttr,
    AttrConstraint,
    BaseAttr,
    ConstraintContext,
    ConstraintVar,
    GenericData,
    MessageConstraint,
    ParamAttrDef,
    ParameterDef,
    irdl_attr_definition,
    irdl_to_attr_constraint,
)
from xdsl.parser import AttrParser
from xdsl.printer import Printer
from xdsl.utils.exceptions import PyRDLAttrDefinitionError, VerifyException


def test_wrong_attribute_type():
    with pytest.raises(
        TypeError,
        match="Class AbstractAttribute should either be a subclass of 'Data' or 'ParametrizedAttribute'",
    ):

        @irdl_attr_definition
        class AbstractAttribute(Attribute):  # pyright: ignore[reportUnusedClass]
            name = "test.wrong"
            pass


################################################################################
# Data attributes
################################################################################


@irdl_attr_definition
class BoolData(Data[bool]):
    """An attribute holding a boolean value."""

    name = "test.bool"

    @classmethod
    def parse_parameter(cls, parser: AttrParser) -> bool:
        with parser.in_angle_brackets():
            if parser.parse_optional_keyword("True"):
                return True
            if parser.parse_optional_keyword("False"):
                return False
            parser.raise_error("Expected True or False literal")

    def print_parameter(self, printer: Printer):
        with printer.in_angle_brackets():
            printer.print_string(str(self.data))


@irdl_attr_definition
class IntData(Data[int]):
    """An attribute holding an integer value."""

    name = "test.int"

    @classmethod
    def parse_parameter(cls, parser: AttrParser) -> int:
        with parser.in_angle_brackets():
            return parser.parse_integer()

    def print_parameter(self, printer: Printer):
        with printer.in_angle_brackets():
            printer.print_string(str(self.data))


@irdl_attr_definition
class StringData(Data[str]):
    """An attribute holding a string value."""

    name = "test.str"

    @classmethod
    def parse_parameter(cls, parser: AttrParser) -> str:
        with parser.in_angle_brackets():
            return parser.parse_str_literal()

    def print_parameter(self, printer: Printer):
        with printer.in_angle_brackets():
            printer.print_string(self.data)


def test_simple_data():
    """Test that the definition of a data with a class parameter."""
    b = BoolData(True)
    stream = StringIO()
    p = Printer(stream=stream)
    p.print_attribute(b)
    assert stream.getvalue() == "#test.bool<True>"


@irdl_attr_definition
class IntListData(Data[tuple[int, ...]]):
    """
    An attribute holding a list of integers.
    """

    name = "test.int_list"

    @classmethod
    def parse_parameter(cls, parser: AttrParser) -> tuple[int]:
        raise NotImplementedError()

    def print_parameter(self, printer: Printer) -> None:
        with printer.in_angle_brackets():
            printer.print_string("[")
            printer.print_list(self.data, lambda x: printer.print_string(str(x)))
            printer.print_string("]")


def test_non_class_data():
    """Test the definition of a Data with a non-class parameter."""
    attr = IntListData((0, 1, 42))
    stream = StringIO()
    p = Printer(stream=stream)
    p.print_attribute(attr)
    assert stream.getvalue() == "#test.int_list<[0, 1, 42]>"


class TestEnum(StrEnum):
    Yes = auto()
    No = auto()
    Maybe = auto()


class TestNonIdentifierEnum(StrEnum):
    """
    The value defined by this StrEnum is not parsable as an identifier, because of the
    contained space.
    While valid as a StrEnum, it is thus invalid to use it in an EnumAttribute.
    """

    Spaced = "left right"


@irdl_attr_definition
class EnumData(EnumAttribute[TestEnum], SpacedOpaqueSyntaxAttribute):
    name = "test.enum"


def test_enum_attribute():
    """Test the definition of an EnumAttribute."""
    attr = EnumData(TestEnum.No)
    assert str(attr) == "#test<enum no>"


def test_indirect_enum_guard():
    EnumType = TypeVar("EnumType", bound=StrEnum)
    with pytest.raises(
        TypeError, match="Only direct inheritance from EnumAttribute is allowed."
    ):

        class IndirectEnumData(  # pyright: ignore[reportUnusedClass]
            EnumAttribute[EnumType]
        ):
            name = "test.indirect_enum"


def test_identifier_enum_guard():
    with pytest.raises(
        ValueError,
        match="All StrEnum values of an EnumAttribute must be parsable as an identifer.",
    ):

        class IndirectEnumData(  # pyright: ignore[reportUnusedClass]
            EnumAttribute[TestNonIdentifierEnum]
        ):
            name = "test.non_identifier_enum"


@irdl_attr_definition
class BitEnumData(BitEnumAttribute[TestEnum]):
    name = "test.bitenum"
    all_value = "all"
    none_value = "none"


@pytest.mark.parametrize(
    "input,output",
    [
        (None, "#test.bitenum<none>"),
        ([], "#test.bitenum<none>"),
        ([TestEnum.No], "#test.bitenum<no>"),
        ([TestEnum.Yes], "#test.bitenum<yes>"),
        ([TestEnum.No, TestEnum.Yes], "#test.bitenum<yes,no>"),
        ([TestEnum.No, TestEnum.Yes, TestEnum.Maybe], "#test.bitenum<all>"),
        ("all", "#test.bitenum<all>"),
        ("none", "#test.bitenum<none>"),
    ],
)
def test_bit_enum_attribute(input: Sequence[TestEnum] | str, output: str):
    attr = BitEnumData(input)
    assert str(attr) == output


def test_bit_enum_invalid_str():
    with pytest.raises(
        TypeError,
        match="expected string parameter to be one of none or all, got helloworld",
    ):
        BitEnumData("helloworld")


################################################################################
# Typed Attribute
################################################################################


def test_typed_attribute():
    with pytest.raises(
        PyRDLAttrDefinitionError,
        match="TypedAttribute TypedAttr should have a 'type' parameter.",
    ):

        @irdl_attr_definition
        class TypedAttr(  # pyright: ignore[reportUnusedClass]
            TypedAttribute
        ):
            name = "test.typed"


################################################################################
# IntegerAttr
################################################################################


def test_signed_integer_attr():
    """Test the verification of a signed integer attribute."""
    with pytest.raises(VerifyException):
        IntegerAttr(1 << 31, IntegerType(32, Signedness.SIGNED))

    with pytest.raises(VerifyException):
        IntegerAttr(-(1 << 31) - 1, IntegerType(32, Signedness.SIGNED))

    IntegerAttr((1 << 31) - 1, IntegerType(32, Signedness.SIGNED))
    IntegerAttr(-(1 << 31), IntegerType(32, Signedness.SIGNED))


def test_unsigned_integer_attr():
    """Test the verification of a unsigned integer attribute."""
    with pytest.raises(VerifyException):
        IntegerAttr(1 << 32, IntegerType(32, Signedness.UNSIGNED))

    with pytest.raises(VerifyException):
        IntegerAttr(-1, IntegerType(32, Signedness.UNSIGNED))

    IntegerAttr((1 << 32) - 1, IntegerType(32, Signedness.UNSIGNED))


def test_signless_integer_attr():
    """Test the verification of a signless integer attribute."""
    with pytest.raises(VerifyException):
        IntegerAttr((1 << 32), IntegerType(32, Signedness.SIGNLESS))

    with pytest.raises(VerifyException):
        IntegerAttr(-(1 << 32) - 1, IntegerType(32, Signedness.SIGNLESS))

    IntegerAttr(1 << 32 - 1, IntegerType(32, Signedness.SIGNLESS))
    IntegerAttr(-(1 << 31), IntegerType(32, Signedness.SIGNLESS))


################################################################################
# PyRDL Base constraints
################################################################################


@irdl_attr_definition
class BoolWrapperAttr(ParametrizedAttribute):
    name = "test.bool_wrapper"

    param: ParameterDef[BoolData]


def test_bose_constraint():
    """Test the verifier of a base attribute type constraint."""
    attr = BoolWrapperAttr((BoolData(True),))
    stream = StringIO()
    p = Printer(stream=stream)
    p.print_attribute(attr)
    assert stream.getvalue() == "#test.bool_wrapper<#test.bool<True>>"


def test_base_constraint_fail():
    """Test the verifier of a union constraint."""
    with pytest.raises(Exception) as e:
        BoolWrapperAttr((StringData("foo"),))
    assert e.value.args[0] == "#test.str<foo> should be of base attribute test.bool"


################################################################################
# PyRDL union constraints
################################################################################


@irdl_attr_definition
class BoolOrIntParamAttr(ParametrizedAttribute):
    name = "test.bool_or_int"

    param: ParameterDef[BoolData | IntData]


def test_union_constraint_left():
    """Test the verifier of a union constraint."""
    attr = BoolOrIntParamAttr((BoolData(True),))
    stream = StringIO()
    p = Printer(stream=stream)
    p.print_attribute(attr)
    assert stream.getvalue() == "#test.bool_or_int<#test.bool<True>>"


def test_union_constraint_right():
    """Test the verifier of a union constraint."""
    attr = BoolOrIntParamAttr((IntData(42),))
    stream = StringIO()
    p = Printer(stream=stream)
    p.print_attribute(attr)
    assert stream.getvalue() == "#test.bool_or_int<#test.int<42>>"


def test_union_constraint_fail():
    """Test the verifier of a union constraint."""
    with pytest.raises(Exception) as e:
        BoolOrIntParamAttr((StringData("foo"),))
    assert e.value.args[0] == "Unexpected attribute #test.str<foo>"


################################################################################
# PyRDL Annotated constraints
################################################################################


class PositiveIntConstr(AttrConstraint):
    def verify(self, attr: Attribute, constraint_context: ConstraintContext) -> None:
        if not isinstance(attr, IntData):
            raise VerifyException(
                f"Expected {IntData.name} attribute, but got {attr.name}."
            )
        if attr.data <= 0:
            raise VerifyException(f"Expected positive integer, got {attr.data}.")


@irdl_attr_definition
class PositiveIntAttr(ParametrizedAttribute):
    name = "test.positive_int"

    param: ParameterDef[Annotated[IntData, PositiveIntConstr()]]


def test_annotated_constraint():
    """Test the verifier of an annotated constraint."""
    attr = PositiveIntAttr((IntData(42),))
    stream = StringIO()
    p = Printer(stream=stream)
    p.print_attribute(attr)
    assert stream.getvalue() == "#test.positive_int<#test.int<42>>"


def test_annotated_constraint_fail():
    """Test that the verifier of an annotated constraint can fail."""
    with pytest.raises(Exception) as e:
        PositiveIntAttr((IntData(-42),))
    assert e.value.args[0] == "Expected positive integer, got -42."


################################################################################
# PyRDL Generic constraints
################################################################################

_T = TypeVar("_T", bound=BoolData | IntData)


@irdl_attr_definition
class ParamWrapperAttr(Generic[_T], ParametrizedAttribute):
    name = "test.int_or_bool_generic"

    param: ParameterDef[_T]

    def __init__(self, param: _T):
        super().__init__((param,))


def test_typevar_attribute_int():
    """Test the verifier of a generic attribute."""
    attr = ParamWrapperAttr[IntData](IntData(42))
    stream = StringIO()
    p = Printer(stream=stream)
    p.print_attribute(attr)
    assert stream.getvalue() == "#test.int_or_bool_generic<#test.int<42>>"


def test_typevar_attribute_bool():
    """Test the verifier of a generic attribute."""
    attr = ParamWrapperAttr[BoolData](BoolData(True))
    stream = StringIO()
    p = Printer(stream=stream)
    p.print_attribute(attr)
    assert stream.getvalue() == "#test.int_or_bool_generic<#test.bool<True>>"


def test_typevar_attribute_fail():
    """Test that the verifier of an generic attribute can fail."""
    with pytest.raises(Exception) as e:
        ParamWrapperAttr(StringData("foo"))  # pyright: ignore
    assert e.value.args[0] == "Unexpected attribute #test.str<foo>"


@irdl_attr_definition
class ParamConstrAttr(ParametrizedAttribute):
    name = "test.param_constr"

    param: ParameterDef[ParamWrapperAttr[IntData]]

    def __init__(self, param: ParameterDef[ParamWrapperAttr[IntData]]):
        super().__init__((param,))


def test_param_attr_constraint():
    """Test the verifier of an attribute with a parametric constraint."""
    attr = ParamConstrAttr(ParamWrapperAttr(IntData(42)))
    stream = StringIO()
    p = Printer(stream=stream)
    p.print_attribute(attr)
    assert (
        stream.getvalue()
        == "#test.param_constr<#test.int_or_bool_generic<#test.int<42>>>"
    )


def test_param_attr_constraint_fail():
    """
    Test that the verifier of an attribute with
    a parametric constraint can fail.
    """
    with pytest.raises(Exception) as e:
        ParamConstrAttr(ParamWrapperAttr(BoolData(True)))  # pyright: ignore
    assert e.value.args[0] == "#test.bool<True> should be of base attribute test.int"


_U = TypeVar("_U", bound=IntData)


@irdl_attr_definition
class NestedParamWrapperAttr(Generic[_U], ParametrizedAttribute):
    name = "test.nested_param_wrapper"

    param: ParameterDef[ParamWrapperAttr[_U]]

    def __init__(self, param: ParameterDef[ParamWrapperAttr[_U]]):
        super().__init__((param,))


def test_nested_generic_constraint():
    """
    Test the verifier of an attribute with a generic
    constraint used in a parametric constraint.
    """
    attr = NestedParamWrapperAttr[IntData](ParamWrapperAttr(IntData(42)))
    stream = StringIO()
    p = Printer(stream=stream)
    p.print_attribute(attr)
    assert (
        stream.getvalue()
        == "#test.nested_param_wrapper<#test.int_or_bool_generic<#test.int<42>>>"
    )


def test_nested_generic_constraint_fail():
    """
    Test that the verifier of an attribute with
    a parametric constraint can fail.
    """
    with pytest.raises(Exception) as e:
        NestedParamWrapperAttr(ParamWrapperAttr(BoolData(True)))  # pyright: ignore
    assert e.value.args[0] == "#test.bool<True> should be of base attribute test.int"


@irdl_attr_definition
class NestedParamConstrAttr(ParametrizedAttribute):
    name = "test.nested_param_constr"

    param: ParameterDef[NestedParamWrapperAttr[Annotated[IntData, PositiveIntConstr()]]]


def test_nested_param_attr_constraint():
    """
    Test the verifier of a nested parametric constraint.
    """
    attr = NestedParamConstrAttr(
        (NestedParamWrapperAttr(ParamWrapperAttr(IntData(42))),)
    )
    stream = StringIO()
    p = Printer(stream=stream)
    p.print_attribute(attr)
    assert (
        stream.getvalue()
        == "#test.nested_param_constr<#test.nested_param_wrapper<#test.int_or_bool_generic<#test.int<42>>>>"
    )


def test_nested_param_attr_constraint_fail():
    """
    Test that the verifier of a nested parametric constraint can fail.
    """
    with pytest.raises(Exception) as e:
        NestedParamConstrAttr((NestedParamWrapperAttr(ParamWrapperAttr(IntData(-42))),))
    assert e.value.args[0] == "Expected positive integer, got -42."


################################################################################
# Message Constraint
################################################################################


@irdl_attr_definition
class InformativeAttr(ParametrizedAttribute):
    name = "test.informative"

    param: ParameterDef[
        Annotated[
            Attribute,
            MessageConstraint(
                NoneAttr,
                "Dear user, here's what this constraint means in your abstraction.",
            ),
        ]
    ]


def test_informative_attribute():
    okay = InformativeAttr((NoneAttr(),))
    okay.verify()

    with pytest.raises(
        VerifyException,
        match="Dear user, here's what this constraint means in your abstraction.\nUnderlying verification failure: #test.int<42> should be of base attribute none",
    ):
        InformativeAttr((IntData(42),))


def test_informative_constraint():
    """
    Test the verifier of an informative constraint.
    """
    constr = MessageConstraint(NoneAttr(), "User-enlightening message.")
    with pytest.raises(
        VerifyException,
        match="User-enlightening message.\nUnderlying verification failure: Expected attribute #none but got #builtin.int<1>",
    ):
        constr.verify(IntAttr(1), ConstraintContext())
    assert constr.can_infer(set())
    assert constr.get_unique_base() == NoneAttr


################################################################################
# GenericData definition
################################################################################

_MissingGenericDataData = TypeVar("_MissingGenericDataData")


@irdl_attr_definition
class MissingGenericDataData(Data[_MissingGenericDataData]):
    name = "test.missing_genericdata"

    @classmethod
    def parse_parameter(cls, parser: AttrParser) -> _MissingGenericDataData:
        raise NotImplementedError()

    def print_parameter(self, printer: Printer) -> None:
        raise NotImplementedError()

    def verify(self) -> None:
        return


class MissingGenericDataDataWrapper(ParametrizedAttribute):
    name = "test.missing_genericdata_wrapper"

    param: ParameterDef[MissingGenericDataData[int]]


def test_data_with_generic_missing_generic_data_failure():
    """
    Test error message when a generic data is used in constraints
    without implementing GenericData.
    """
    with pytest.raises(Exception) as e:
        irdl_attr_definition(MissingGenericDataDataWrapper)
    assert e.value.args[0] == (
        "Generic `Data` type 'test.missing_genericdata' cannot be converted to "
        "an attribute constraint. Consider making it inherit from "
        "`GenericData` instead of `Data`."
    )


A = TypeVar("A", bound=Attribute)


@dataclass(frozen=True)
class DataListAttr(AttrConstraint):
    """
    A constraint that enforces that the elements of a ListData all respect
    a constraint.
    """

    elem_constr: AttrConstraint

    def verify(
        self,
        attr: Attribute,
        constraint_context: ConstraintContext,
    ) -> None:
        attr = cast(ListData[Attribute], attr)
        for e in attr.data:
            self.elem_constr.verify(e, constraint_context)


@irdl_attr_definition
class ListData(GenericData[list[A]]):
    name = "test.list"

    @classmethod
    def parse_parameter(cls, parser: AttrParser) -> list[A]:
        raise NotImplementedError()

    def print_parameter(self, printer: Printer) -> None:
        with printer.in_angle_brackets():
            printer.print_string("[")
            printer.print_list(self.data, printer.print_attribute)
            printer.print_string("]")

    @staticmethod
    def generic_constraint_coercion(args: tuple[Any]) -> AttrConstraint:
        assert len(args) == 1
        return DataListAttr(irdl_to_attr_constraint(args[0]))

    @staticmethod
    def from_list(data: list[A]) -> ListData[A]:
        return ListData(data)


AnyListData: TypeAlias = ListData[Attribute]


class Test_generic_data_verifier:
    def test_generic_data_verifier(self):
        """
        Test that a GenericData can be created.
        """
        attr = ListData([BoolData(True), ListData([BoolData(False)])])
        stream = StringIO()
        p = Printer(stream=stream)
        p.print_attribute(attr)
        assert (
            stream.getvalue()
            == "#test.list<[#test.bool<True>, #test.list<[#test.bool<False>]>]>"
        )


@irdl_attr_definition
class ListDataWrapper(ParametrizedAttribute):
    name = "test.list_wrapper"

    val: ParameterDef[ListData[BoolData]]


def test_generic_data_wrapper_verifier():
    """
    Test that a GenericData used in constraints pass the verifier when correct.
    """
    attr = ListDataWrapper((ListData([BoolData(True), BoolData(False)]),))
    stream = StringIO()
    p = Printer(stream=stream)
    p.print_attribute(attr)
    assert (
        stream.getvalue()
        == "#test.list_wrapper<#test.list<[#test.bool<True>, #test.bool<False>]>>"
    )


def test_generic_data_wrapper_verifier_failure():
    """
    Test that a GenericData used in constraints fails
    the verifier when constraints are not satisfied.
    """
    with pytest.raises(VerifyException) as e:
        ListDataWrapper((ListData([BoolData(True), ListData([BoolData(False)])]),))
    assert (
        e.value.args[0]
        == "#test.list<[#test.bool<False>]> should be of base attribute test.bool"
    )
    assert (
        e.value.args[0]
        == "#test.list<[#test.bool<False>]> should be of base attribute test.bool"
    )


@irdl_attr_definition
class ListDataNoGenericsWrapper(ParametrizedAttribute):
    name = "test.list_no_generics_wrapper"

    val: ParameterDef[AnyListData]


def test_generic_data_no_generics_wrapper_verifier():
    """
    Test that GenericType can be used in constraints without a parameter.
    """
    attr = ListDataNoGenericsWrapper(
        (ListData([BoolData(True), ListData([BoolData(False)])]),)
    )
    stream = StringIO()
    p = Printer(stream=stream)
    p.print_attribute(attr)
    assert (
        stream.getvalue()
        == "#test.list_no_generics_wrapper<#test.list<[#test.bool<True>, #test.list<[#test.bool<False>]>]>>"
    )


################################################################################
# Parametrized attribute definition
################################################################################


@irdl_attr_definition
class ParamAttrDefAttr(ParametrizedAttribute):
    name = "test.param_attr_def_attr"

    arg1: ParameterDef[Attribute]
    arg2: ParameterDef[BoolData]

    # Check that we can define methods in attribute definition
    def test(self):
        pass


def test_irdl_definition():
    """Test that we can get the IRDL definition of a parametrized attribute."""

    assert ParamAttrDefAttr.get_irdl_definition() == ParamAttrDef(
        "test.param_attr_def_attr", [("arg1", AnyAttr()), ("arg2", BaseAttr(BoolData))]
    )


class InvalidTypedFieldTestAttr(ParametrizedAttribute):
    name = "test.invalid_typed_field"

    x: int


def test_invalid_typed_field():
    """Check that typed fields are not allowed."""
    with pytest.raises(PyRDLAttrDefinitionError):
        irdl_attr_definition(InvalidTypedFieldTestAttr)


class InvalidUntypedFieldTestAttr(ParametrizedAttribute):
    name = "test.invalid_untyped_field"

    x = 2


def test_invalid_field():
    """Check that untyped fields are not allowed."""
    with pytest.raises(PyRDLAttrDefinitionError):
        irdl_attr_definition(InvalidUntypedFieldTestAttr)


@irdl_attr_definition
class OveriddenInitAttr(ParametrizedAttribute):
    name = "test.overidden_init"

    param: ParameterDef[Attribute]

    def __init__(self, param: int | str):
        match param:
            case int():
                super().__init__((IntData(param),))
            case str():
                super().__init__((StringData(param),))


def test_generic_constructor():
    """Test the generic constructor of a ParametrizedAttribute."""

    param = IntData(42)
    attr = OveriddenInitAttr.new([param])

    assert isinstance(attr, OveriddenInitAttr)
    assert attr.param == param


def test_custom_constructor():
    """Test the use of custom constructors in ParametrizedAttribute."""

    assert OveriddenInitAttr.new([IntData(42)]) == OveriddenInitAttr(42)
    assert OveriddenInitAttr.new([StringData("17")]) == OveriddenInitAttr("17")


################################################################################
# ConstraintVar
################################################################################


@irdl_attr_definition
class ConstraintVarAttr(ParametrizedAttribute):
    name = "test.constraint_var"

    T = Annotated[IntegerType, ConstraintVar("T")]

    param1: ParameterDef[IntegerAttr[T]]
    param2: ParameterDef[IntegerAttr[T]]


def test_constraint_var():
    """Test that ConstraintVar can be used in attributes."""
    ConstraintVarAttr.new([IntegerAttr(42, 32), IntegerAttr(17, 32)])
    ConstraintVarAttr.new([IntegerAttr(42, 64), IntegerAttr(17, 64)])


def test_constraint_var_fail_non_equal():
    """Test that constraint variables must be equal."""
    with pytest.raises(VerifyException):
        ConstraintVarAttr.new([IntegerAttr(42, 32), IntegerAttr(42, 64)])
    with pytest.raises(VerifyException):
        ConstraintVarAttr.new([IntegerAttr(42, 64), IntegerAttr(42, 32)])


def test_constraint_var_fail_not_satisfy_constraint():
    """Test that constraint variables must satisfy the underlying constraint."""
    with pytest.raises(VerifyException):
        ConstraintVarAttr.new(
            [IntegerAttr(42, IndexType()), IntegerAttr(17, IndexType())]
        )



"""
Test HW traits and interfaces.
"""

from unittest.mock import ANY, patch

import pytest

from xdsl.context import Context
from xdsl.dialects.builtin import ArrayAttr, StringAttr, SymbolRefAttr, i32, i64
from xdsl.dialects.hw import (
    HW,
    Direction,
    DirectionAttr,
    HWModuleLike,
    HWModuleOp,
    InnerRefAttr,
    InnerRefNamespaceTrait,
    InnerRefUserOpInterfaceTrait,
    InnerSymAttr,
    InnerSymbolTableCollection,
    InnerSymbolTableTrait,
    InnerSymPropertiesAttr,
    InnerSymTarget,
    InstanceOp,
    ModulePort,
    ModuleType,
)
from xdsl.dialects.test import TestOp
from xdsl.ir import Block
from xdsl.irdl import (
    IRDLOperation,
    Region,
    attr_def,
    irdl_op_definition,
    opt_region_def,
    region_def,
    traits_def,
)
from xdsl.parser import Parser
from xdsl.traits import (
    IsTerminator,
    SingleBlockImplicitTerminator,
    SymbolOpInterface,
    SymbolTable,
    ensure_terminator,
)
from xdsl.utils.exceptions import VerifyException
from xdsl.utils.test_value import TestSSAValue


def test_inner_sym_target():
    invalid_target = InnerSymTarget()
    assert not invalid_target

    operand1 = TestOp()

    target = InnerSymTarget(operand1)
    assert target
    assert target.is_op_only()
    assert not target.is_field()

    sub_target = InnerSymTarget.get_target_for_subfield(target, 1)
    assert isinstance(sub_target, InnerSymTarget)
    assert sub_target
    assert not sub_target.is_op_only()
    assert sub_target.is_field()


# Module / Circuit / Wire / Output naming taken from FIRRTL classes using these interfaces.
@irdl_op_definition
class ModuleOp(IRDLOperation):
    name = "module"
    region = region_def()
    sym_name = attr_def(StringAttr)
    traits = traits_def(InnerSymbolTableTrait(), SymbolOpInterface())


@irdl_op_definition
class OutputOp(IRDLOperation):
    name = "output"
    traits = traits_def(IsTerminator())


@irdl_op_definition
class CircuitOp(IRDLOperation):
    name = "circuit"
    region: Region | None = opt_region_def()
    sym_name = attr_def(StringAttr)
    traits = traits_def(
        InnerRefNamespaceTrait(),
        SymbolTable(),
        SingleBlockImplicitTerminator(OutputOp),
    )

    def __post_init__(self):
        for trait in self.get_traits_of_type(SingleBlockImplicitTerminator):
            ensure_terminator(self, trait)


@irdl_op_definition
class WireOp(IRDLOperation):
    name = "wire"
    sym_name = attr_def(StringAttr)
    traits = traits_def(InnerRefUserOpInterfaceTrait())


def test_inner_symbol_table_interface():
    """
    Test operations that conform to InnerSymbolTableTrait
    """
    mod = ModuleOp(
        attributes={"sym_name": StringAttr("symbol_name")}, regions=[[OutputOp()]]
    )

    circ = CircuitOp(attributes={"sym_name": StringAttr("other_name")}, regions=[[mod]])
    circ.verify()

    mod_no_parent = ModuleOp(
        attributes={"sym_name": StringAttr("symbol_name")}, regions=[[OutputOp()]]
    )
    with pytest.raises(
        VerifyException,
        match="Operation module with trait InnerSymbolTableTrait must have a parent with trait SymbolOpInterface",
    ):
        mod_no_parent.verify()

    mod_no_trait_circ = ModuleOp(
        attributes={"sym_name": StringAttr("symbol_name")}, regions=[[OutputOp()]]
    )
    no_trait_circ = TestOp(regions=[[mod_no_trait_circ, OutputOp()]])
    with pytest.raises(
        VerifyException,
        match="Operation module with trait InnerSymbolTableTrait must have a parent with trait InnerRefNamespaceLike",
    ):
        mod_no_trait_circ.verify()
    with pytest.raises(
        VerifyException,
        match="Operation module with trait InnerSymbolTableTrait must have a parent with trait InnerRefNamespaceLike",
    ):
        no_trait_circ.verify()

    @irdl_op_definition
    class MissingTraitModuleOp(IRDLOperation):
        name = "module"
        region = region_def()
        sym_name = attr_def(StringAttr)
        traits = traits_def(InnerSymbolTableTrait())

    mod_missing_trait = MissingTraitModuleOp(
        attributes={"sym_name": StringAttr("symbol_name")}, regions=[[OutputOp()]]
    )
    circ_mod_missing_trait = CircuitOp(regions=[[mod_missing_trait, OutputOp()]])
    with pytest.raises(
        VerifyException, match="Operation module must have trait SymbolOpInterface"
    ):
        mod_missing_trait.verify()
    with pytest.raises(
        VerifyException, match="Operation module must have trait SymbolOpInterface"
    ):
        circ_mod_missing_trait.verify()

    @irdl_op_definition
    class MissingAttrModuleOp(IRDLOperation):
        name = "module"
        region = region_def()
        traits = traits_def(InnerSymbolTableTrait(), SymbolOpInterface())

    mod_missing_trait_parent = ModuleOp(regions=[[OutputOp()]])
    MissingAttrModuleOp(regions=[[mod_missing_trait_parent, OutputOp()]])
    with pytest.raises(
        VerifyException,
        match="attribute sym_name expected",
    ):
        mod_missing_trait_parent.verify()


def test_inner_ref_namespace_interface():
    """
    Test operations that conform to InnerRefNamespaceTrait
    """

    @irdl_op_definition
    class MissingTraitCircuitOp(IRDLOperation):
        name = "circuit"
        region: Region | None = opt_region_def()
        sym_name = attr_def(StringAttr)
        traits = traits_def(
            InnerRefNamespaceTrait(), SingleBlockImplicitTerminator(OutputOp)
        )

    wire0 = WireOp(attributes={"sym_name": StringAttr("wire0")})
    mod0 = MissingTraitCircuitOp(
        attributes={"sym_name": StringAttr("mod0")}, regions=[[wire0, OutputOp()]]
    )
    circuit_no_symboltable = CircuitOp(
        attributes={"sym_name": StringAttr("circuit")}, regions=[[mod0]]
    )

    with pytest.raises(
        VerifyException, match="Operation circuit must have trait SymbolTable"
    ):
        circuit_no_symboltable.verify()

    wire1 = WireOp(attributes={"sym_name": StringAttr("wire1")})
    wire2 = WireOp(attributes={"sym_name": StringAttr("wire2")})
    mod1 = ModuleOp(
        attributes={"sym_name": StringAttr("mod1")}, regions=[[wire1, OutputOp()]]
    )
    mod2 = ModuleOp(
        attributes={"sym_name": StringAttr("mod2")}, regions=[[wire2, OutputOp()]]
    )
    circuit = CircuitOp(
        attributes={"sym_name": StringAttr("circuit")}, regions=[[mod1, mod2]]
    )

    # InnerRefUserOpInterfaceTrait.verify_inner_refs() does not do anything, so just mock
    # to check it is called
    with patch.object(
        InnerRefUserOpInterfaceTrait, "verify_inner_refs"
    ) as inner_ref_verif:
        circuit.verify()

    inner_ref_verif.assert_any_call(wire1, ANY)
    inner_ref_verif.assert_any_call(wire2, ANY)
    assert inner_ref_verif.call_count == 2


def test_inner_symbol_table_collection():
    """
    Test operations that pertain to InnerSymbolTableCollection
    """
    wire1 = WireOp(attributes={"sym_name": StringAttr("wire1")})
    wire2 = WireOp(attributes={"sym_name": StringAttr("wire2")})
    mod1 = ModuleOp(
        attributes={"sym_name": StringAttr("mod1")}, regions=[[wire1, OutputOp()]]
    )
    mod2 = ModuleOp(
        attributes={"sym_name": StringAttr("mod2")}, regions=[[wire2, OutputOp()]]
    )
    circuit = CircuitOp(
        attributes={"sym_name": StringAttr("circuit")}, regions=[[mod1, mod2]]
    )

    with pytest.raises(
        VerifyException, match="Operation wire should have InnerRefNamespaceTrait trait"
    ):
        inner_sym_tables = InnerSymbolTableCollection(wire1)

    inner_sym_tables = InnerSymbolTableCollection(circuit)

    with pytest.raises(
        VerifyException, match=r"Trying to insert the same op twice in symbol tables:"
    ):
        inner_sym_tables.populate_and_verify_tables(circuit)

    with pytest.raises(
        VerifyException, match="Operation wire should have InnerSymbolTableTrait trait"
    ):
        inner_sym_tables.get_inner_symbol_table(wire1)

    sym_table1 = inner_sym_tables.get_inner_symbol_table(mod1)
    sym_table2 = inner_sym_tables.get_inner_symbol_table(mod2)
    assert sym_table1 is not sym_table2, (
        "Different InnerSymbolTableTrait objects must return different instances of inner symbol tables"
    )

    unpopulated_inner_sym_tables = InnerSymbolTableCollection()
    sym_table3 = unpopulated_inner_sym_tables.get_inner_symbol_table(mod1)
    sym_table4 = unpopulated_inner_sym_tables.get_inner_symbol_table(mod2)
    assert sym_table3 is not sym_table4, (
        "InnerSymbolTableTrait still behave as expected when created on the fly"
    )


def test_inner_ref_attr():
    """
    Test inner reference attributes
    """
    wire1 = WireOp(attributes={"sym_name": StringAttr("wire1")})
    wire2 = WireOp(attributes={"sym_name": StringAttr("wire2")})
    mod1 = ModuleOp(
        attributes={"sym_name": StringAttr("mod1")}, regions=[[wire1, OutputOp()]]
    )
    mod2 = ModuleOp(
        attributes={"sym_name": StringAttr("mod2")}, regions=[[wire2, OutputOp()]]
    )
    CircuitOp(attributes={"sym_name": StringAttr("circuit")}, regions=[[mod1, mod2]])

    ref = InnerRefAttr("mod2", "wire2")
    assert ref.get_module().data == "mod2", (
        "Name of the referenced module should be returned correctly"
    )


def test_inner_sym_attr():
    """
    Test inner symbol attributes
    """
    invalid_sym_attr = InnerSymAttr()
    assert invalid_sym_attr.get_sym_name() is None, (
        "Invalid InnerSymAttr should return no name"
    )

    sym_attr = InnerSymAttr("sym")
    assert sym_attr.get_sym_name() == StringAttr("sym"), (
        "InnerSymAttr for “ground” type should return name"
    )

    with pytest.raises(VerifyException, match=r"inner symbol cannot have empty name"):
        InnerSymAttr("")

    aggregate_sym_attr = InnerSymAttr(
        [
            InnerSymPropertiesAttr("sym", 0, "public"),
            InnerSymPropertiesAttr("other", 1, "private"),
            InnerSymPropertiesAttr("yet_another", 2, "nested"),
        ]
    )

    assert aggregate_sym_attr.get_sym_name() == StringAttr("sym"), (
        "InnerSymAttr for aggregate types should return name with field ID 0"
    )

    for inner, expected_field_id in zip(aggregate_sym_attr, [0, 1, 2]):
        assert inner.field_id.data == expected_field_id, (
            "InnerSymAttr should allow iterating its properties in order"
        )

    aggregate_without_nested = aggregate_sym_attr.erase(2)
    assert aggregate_without_nested.get_sym_if_exists(2) is None, (
        "InnerSymAttr removal should work"
    )
    assert len(aggregate_without_nested) == 2, (
        "InnerSymAttr removal should correctly change length"
    )


def test_instance_builder():
    MODULE_CTX = """
hw.module @module(in %foo: i32, in %bar: i64, out baz: i32, out qux: i64) {
  hw.output %foo, %bar : i32, i64
}
"""

    ctx = Context()
    ctx.load_dialect(HW)

    module_op = Parser(ctx, MODULE_CTX).parse_module()

    module_op.body.block.add_op(
        inst_op := InstanceOp(
            "test",
            SymbolRefAttr("module"),
            (("foo", TestSSAValue(i32)), ("bar", TestSSAValue(i64))),
            (("baz", i32), ("qux", i64)),
        )
    )

    inst_op.verify()
    assert inst_op.instance_name == StringAttr("test")
    assert inst_op.module_name == SymbolRefAttr("module")
    assert inst_op.arg_names.data == (StringAttr("foo"), StringAttr("bar"))
    assert inst_op.result_names.data == (StringAttr("baz"), StringAttr("qux"))

    assert inst_op.operand_types == (i32, i64)
    assert inst_op.result_types == (i32, i64)


def test_hwmoduleop_hwmodulelike():
    module_type = ModuleType((ArrayAttr(()),))

    hw_module = HWModuleOp(
        StringAttr("foo"), module_type, Region((Block((OutputOp(),)),))
    )

    hw_module_like = hw_module.get_trait(HWModuleLike)
    assert hw_module_like is not None
    assert hw_module_like.get_hw_module_type(hw_module) == module_type

    new_module_type = ModuleType(
        (
            ArrayAttr(
                (ModulePort((StringAttr("in1"), i32, DirectionAttr(Direction.INPUT))),)
            ),
        )
    )
    hw_module_like.set_hw_module_type(hw_module, new_module_type)
    assert hw_module_like.get_hw_module_type(hw_module) == new_module_type



"""
Generic implementation of a disjoint set data structure.

See external [documentation](https://en.wikipedia.org/wiki/Disjoint-set_data_structure).
"""

from collections.abc import Hashable, Sequence
from typing import Generic, TypeVar


class IntDisjointSet:
    """
    Represents a collection of disjoint sets of integers.
    The integers stored are always in the range [0,n), where n is the number of elements
    in this structure.

    This implementation uses path compression and union by size for efficiency.
    The amortized time complexity for operations is nearly constant.
    """

    _parent: list[int]
    """
    Index of the parent node. If the node is its own parent then it is a root node.
    """
    _count: list[int]
    """
    If the node is a root node, the corresponding value is the count of elements in the
    set. For non-root nodes, these counts may be stale and should not be used.
    """

    def __init__(self, *, size: int) -> None:
        """
        Initialize disjoint sets with elements [0,size).
        Each element starts in its own singleton set.
        """
        self._parent = list(range(size))
        self._count = [1] * size

    def value_count(self) -> int:
        """Number of nodes in this structure."""
        return len(self._parent)

    def add(self) -> int:
        """
        Add a new element to this set as a singleton.
        Returns the added value, which will be equal to the previous size.
        """
        res = len(self._parent)
        self._parent.append(res)
        self._count.append(1)
        return res

    def __getitem__(self, value: int) -> int:
        """
        Returns the root/representative value of this set.
        Uses path compression - updates parent pointers to point directly to the root
        as we traverse up the tree, improving amortized performance.
        """
        if value < 0 or len(self._parent) <= value:
            raise KeyError(f"Index {value} not found")

        # Find the root
        root = value
        while self._parent[root] != root:
            root = self._parent[root]

        # Path compression - point all nodes on path to root
        current = value
        while current != root:
            next_parent = self._parent[current]
            self._parent[current] = root
            current = next_parent

        return root

    def union(self, lhs: int, rhs: int) -> bool:
        """
        Merges the sets containing lhs and rhs if they are different.
        Returns True if the sets were merged, False if they were already the same set.

        Uses union by size - the smaller tree is attached to the larger tree's root
        to maintain balance. This ensures the maximum tree height is O(log n).
        """
        lhs_root = self[lhs]
        rhs_root = self[rhs]
        if lhs_root == rhs_root:
            return False

        lhs_count = self._count[lhs_root]
        rhs_count = self._count[rhs_root]
        # Choose the root of the larger tree as the new parent
        new_parent, new_child = (
            (lhs_root, rhs_root) if lhs_count <= rhs_count else (rhs_root, lhs_root)
        )
        self._parent[new_child] = new_parent
        self._count[new_parent] = lhs_count + rhs_count
        # Note: We don't need to update _count[new_child] since it's no longer a root
        return True

    def connected(self, lhs: int, rhs: int) -> bool:
        return self[lhs] == self[rhs]


_T = TypeVar("_T", bound=Hashable)


class DisjointSet(Generic[_T]):
    """
    A disjoint-set data structure that works with arbitrary hashable values.
    Internally uses IntDisjointSet by mapping values to integer indices.
    """

    _base: IntDisjointSet
    _values: list[_T]
    _index_by_value: dict[_T, int]

    def __init__(self, values: Sequence[_T] = ()):
        """
        Initialize a DisjointSet with the given sequence of values.
        Each value starts in its own singleton set.

        Args:
            values: Initial sequence of values to add to the disjoint set
        """
        self._values = list(values)
        self._index_by_value = {v: i for i, v in enumerate(self._values)}
        self._base = IntDisjointSet(size=len(self._values))

    def __len__(self):
        return len(self._values)

    def add(self, value: _T):
        """
        Add a new value to the disjoint set in its own singleton set.

        Args:
            value: The value to add
        """
        index = self._base.add()
        self._values.append(value)
        self._index_by_value[value] = index

    def find(self, value: _T) -> _T:
        """
        Find the representative value for the set containing the given value.

        Returns the representative value for the set.

        Raises:
            KeyError: If the value is not in the disjoint set
        """
        index = self._base[self._index_by_value[value]]
        return self._values[index]

    def union(self, lhs: _T, rhs: _T) -> bool:
        """
        Merge the sets containing the two given values if they are different.

        Returns `True` if the sets were merged, `False` if they were already the same
        set.

        Raises:
            KeyError: If either value is not in the disjoint set
        """
        return self._base.union(self._index_by_value[lhs], self._index_by_value[rhs])

    def connected(self, lhs: _T, rhs: _T) -> bool:
        """
        Returns `True` if the values are in the same set.

        Raises:
            KeyError: If either value is not in the disjoint set
        """
        return self._base.connected(
            self._index_by_value[lhs], self._index_by_value[rhs]
        )



from dataclasses import dataclass, field
from itertools import repeat

from xdsl.builder import ImplicitBuilder
from xdsl.context import Context
from xdsl.dialects import memref_stream
from xdsl.dialects.builtin import (
    AffineMapAttr,
    ArrayAttr,
    IntegerAttr,
    ModuleOp,
)
from xdsl.ir import Block, Region, SSAValue
from xdsl.ir.affine import AffineExpr
from xdsl.passes import ModulePass
from xdsl.pattern_rewriter import (
    PatternRewriter,
    PatternRewriteWalker,
    RewritePattern,
    op_type_rewrite_pattern,
)


@dataclass(frozen=True)
class PipelineGenericPattern(RewritePattern):
    pipeline_depth: int = field()

    @op_type_rewrite_pattern
    def match_and_rewrite(
        self, op: memref_stream.GenericOp, rewriter: PatternRewriter
    ) -> None:
        if memref_stream.IteratorTypeAttr.interleaved() in op.iterator_types:
            # Already interleaved
            return

        if memref_stream.IteratorTypeAttr.reduction() not in op.iterator_types:
            # No reduction
            return

        interleave_bound_index = -1
        interleave_bound = -1
        for index, (iterator_type, bound) in enumerate(
            zip(op.iterator_types, op.bounds, strict=True)
        ):
            if iterator_type == memref_stream.IteratorTypeAttr.parallel():
                interleave_bound = bound.value.data
                interleave_bound_index = index
        if interleave_bound == -1:
            # No parallel dimension
            return

        interleave_factor = 1
        # Search factors until the next number divisible by pipeline_depth
        for potential_factor in range(1, self.pipeline_depth * 2):
            if not interleave_bound % potential_factor:
                # Found a larger factor
                interleave_factor = potential_factor

        old_block = op.body.block
        new_region = Region(
            Block(
                arg_types=(
                    t
                    for arg in old_block.args
                    for t in repeat(arg.type, interleave_factor)
                )
            )
        )
        with ImplicitBuilder(new_region) as args:
            # For each interleaved block replica, a mapping from old values to new values
            value_map: tuple[dict[SSAValue, SSAValue], ...] = tuple(
                {} for _ in range(interleave_factor)
            )
            for arg_index, new_arg in enumerate(args):
                old_arg = old_block.args[arg_index // interleave_factor]
                value_map[arg_index % interleave_factor][old_arg] = new_arg
                new_arg.name_hint = old_arg.name_hint
            for block_op in old_block.ops:
                if isinstance(block_op, memref_stream.YieldOp):
                    memref_stream.YieldOp(
                        *([vm[arg] for vm in value_map for arg in block_op.arguments])
                    )
                else:
                    for i in range(interleave_factor):
                        block_op.clone(value_mapper=value_map[i])

        # New maps are the same, except that they have one more dimension and the
        # dimension that is interleaved is updated to
        # `dim * interleave_factor + new_dim`.
        new_indexing_maps = ArrayAttr(
            AffineMapAttr(
                m.data.replace_dims_and_symbols(
                    (
                        tuple(
                            AffineExpr.dimension(i)
                            for i in range(interleave_bound_index)
                        )
                        + (
                            AffineExpr.dimension(interleave_bound_index)
                            * interleave_factor
                            + AffineExpr.dimension(m.data.num_dims),
                        )
                        + tuple(
                            AffineExpr.dimension(i)
                            for i in range(
                                interleave_bound_index + 1, m.data.num_dims + 2
                            )
                        )
                    ),
                    (),
                    m.data.num_dims + 1,
                    0,
                )
            )
            for m in op.indexing_maps
        )

        # The new bounds are the same, except there is one more bound
        new_bounds = list(op.bounds)
        new_bounds.append(IntegerAttr.from_index_int_value(interleave_factor))
        new_bounds[interleave_bound_index] = IntegerAttr.from_index_int_value(
            interleave_bound // interleave_factor
        )

        rewriter.replace_matched_op(
            memref_stream.GenericOp(
                op.inputs,
                op.outputs,
                op.inits,
                new_region,
                new_indexing_maps,
                ArrayAttr(
                    op.iterator_types.data
                    + (memref_stream.IteratorTypeAttr.interleaved(),)
                ),
                ArrayAttr(new_bounds),
                op.init_indices,
                op.doc,
                op.library_call,
            )
        )


@dataclass(frozen=True)
class MemRefStreamInterleavePass(ModulePass):
    """
    Tiles the innermost parallel dimension of a `memref_stream.generic`.
    If specified, the `pipeline-depth` parameter specifies the number of operations in the
    resulting body that should be executed concurrently.
    The pass will select the largest factor of the corresponding bound smaller than
    `pipeline-depth * 2`.
    The search range is bound by `pipeline-depth * 2` as very large interleaving factors
    can increase register pressure and potentially exhaust all available registers.
    In the future, it would be good to take the number of available registers into account
    when choosing a search range, as well as inspecting the generic body for
    read-after-write dependencies.
    """

    name = "memref-stream-interleave"

    pipeline_depth: int = field(default=4)

    def apply(self, ctx: Context, op: ModuleOp) -> None:
        PatternRewriteWalker(
            PipelineGenericPattern(self.pipeline_depth),
            apply_recursively=False,
        ).rewrite_module(op)
